"""
Dereference and format pointer chains.
"""

from __future__ import annotations

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
import pwndbg.color.memory as mem_color
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.enhance
from pwndbg.color import ColorConfig
from pwndbg.color import ColorParamSpec
from pwndbg.color import theme

LIMIT = pwndbg.config.add_param(
    "dereference-limit", 5, "max number of pointers to dereference in a chain"
)

c = ColorConfig(
    "chain",
    [
        ColorParamSpec("arrow", "normal", "color of chain formatting (arrow)"),
        ColorParamSpec(
            "contiguous-marker", "normal", "color of chain formatting (contiguous marker)"
        ),
    ],
)


def get(
    address: int | None,
    limit: int = LIMIT,
    offset: int = 0,
    hard_stop: int | None = None,
    hard_end: int = 0,
    include_start: bool = True,
    safe_linking: bool = False,
) -> list[int] | None:
    """
    Recursively dereferences an address. For bare metal, it will stop when the address is not in any of vmmap pages to avoid redundant dereference.

    Arguments:
        address: the first address to begin dereferencing
        limit: number of valid pointers
        offset: offset into the address to get the next pointer
        hard_stop: address to stop at
        hard_end: value to append when hard_stop is reached
        include_start: whether to include starting address or not
        safe_linking: whether this chain use safe-linking

    Returns:
        A list representing pointers of each ```address``` and reference
    """
    if address is None:
        return None
    assert address >= 0, "address must be positive"

    limit = int(limit)

    result = [address] if include_start else []

    is_pagefault_supported = pwndbg.aglib.memory.is_pagefault_supported()

    for _ in range(limit):
        # Don't follow cycles, except to stop at the second occurrence.
        if result.count(address) >= 2:
            break

        if hard_stop is not None and address == hard_stop:
            result.append(hard_end)
            break

        try:
            address = address + offset

            # On embedded systems, it's non uncommon for MMIO regions to exist where memory reads might mutate the hardware/process state.
            # This check prevents the memory dereferences to protect against this case.
            # See discussion here: https://github.com/pwndbg/pwndbg/pull/385
            if not is_pagefault_supported and not pwndbg.aglib.vmmap.find(address):
                break

            next_address = pwndbg.aglib.memory.read_pointer_width(address)
            address = next_address ^ ((address >> 12) if safe_linking else 0)
            address &= pwndbg.aglib.arch.ptrmask
            result.append(address)
        except pwndbg.dbg_mod.Error:
            break

    return result


config_arrow_left = theme.add_param("chain-arrow-left", "◂—", "left arrow of chain formatting")
config_arrow_right = theme.add_param("chain-arrow-right", "—▸", "right arrow of chain formatting")
config_contiguous = theme.add_param(
    "chain-contiguous-marker", "...", "contiguous marker of chain formatting"
)


def format(
    value: int | list[int] | None,
    limit: int = LIMIT,
    code: bool = True,
    offset: int = 0,
    hard_stop: int | None = None,
    hard_end: int = 0,
    safe_linking: bool = False,
    enhance_string_len: int | None = None,
) -> str:
    """
    Recursively dereferences an address into string representation, or convert the list representation
    of address dereferences into string representation.

    Arguments:
        value: Either the starting address to be sent to get, or the result of get (a list)
        limit: Number of valid pointers
        code: Hint that indicates the value may be an instruction
        offset: Offset into the address to get the next pointer
        hard_stop: Value to stop on
        hard_end: Value to append when hard_stop is reached: null, value of hard stop, a string.
        safe_linking: whether this chain use safe-linking
        enhance_string_len: The length of string to display for enhancement of the last pointer
    Returns:
        A string representing pointers of each address and reference
        Strings format: 0x0804a10 —▸ 0x08061000 ◂— 0x41414141
    """
    if value is None:
        return "<unavailable>"

    limit = int(limit)

    # Allow results from get function to be passed to format
    if isinstance(value, list):
        chain = value
    else:
        chain = get(value, limit, offset, hard_stop, hard_end, safe_linking=safe_linking) or []

    arrow_left = c.arrow(f" {config_arrow_left} ")
    arrow_right = c.arrow(f" {config_arrow_right} ")

    # Ask the decompiler to resolve stack variables
    stack_vars = pwndbg.dintegration.manager.get_stack_var_dict_all()

    # Colorize the chain
    rest = [
        mem_color.get_address_and_symbol(addr, stack_vars) if addr >= 0 else "" for addr in chain
    ]

    # If the dereference limit is zero, skip any enhancements.
    if limit == 0:
        return rest[0]
    # Otherwise replace last element with the enhanced information.
    rest = rest[:-1]

    # Enhance the last entry
    # If there are no pointers (e.g. eax = 0x41414141), then enhance
    # the only element there is.
    if len(chain) == 1:
        # Note the "attempt_dereference" argument, which is set to False.
        # In general, this function assumes that the caller has manually fully dereferenced the input list of pointers.
        # If the only value in the list is a pointer, the function assumes this is purposeful and that that pointer cannot be dereferenced.
        # This is because the code that generated the list determined that we cannot safely reason about the dereferenced value at the current program state.
        # This case only applies to lists of length one, because if the list has more than one value, we already know
        # that the second to last value, chain[-2], can be safely dereferenced - how else would chain[-1] exist?
        # In other case where chain[-1] is not a pointer, the argument has no effect.
        enhanced = pwndbg.enhance.enhance(
            chain[-1],
            code=code,
            attempt_dereference=False,
            enhance_string_len=enhance_string_len,
        )
    # We want to enhance the last pointer value. If an offset was used
    # chain failed at that offset, so display that offset.
    elif len(chain) < limit + 1:
        enhanced = pwndbg.enhance.enhance(
            chain[-2] + offset,
            code=code,
            safe_linking=safe_linking,
            enhance_string_len=enhance_string_len,
        )

    else:
        enhanced = c.contiguous_marker(f"{config_contiguous}")

    if len(chain) == 1:
        return enhanced

    return arrow_right.join(rest) + arrow_left + enhanced



"""
Rizin integration with rzpipe.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pwndbg.aglib.elf
import pwndbg.aglib.kernel
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.lib.cache
from pwndbg.color import message
from pwndbg.lib import ErrorCode
from pwndbg.lib import Status

if TYPE_CHECKING:
    import rzpipe as rzpipe_mod


@pwndbg.lib.cache.cache_until("start", "objfile")
def rzpipe() -> rzpipe_mod.open | Status:
    """
    Spawn and return a rzpipe handle for the current process file.
    This function requires a rizin installation plus the rzpipe python
    library. The base address is automatically set for PIE when loading the
    binary.
    After opening the handle, the binary is automatically analyzed.

    If successful, returns the handle, otherwise returns a failing Status.
    """
    filename = pwndbg.aglib.proc.exe()
    if filename is None:
        return Status.fail("Could not find objfile to create a rzpipe for")

    try:
        import rzpipe as rzpipe_mod
    except ImportError:
        return Status.coded_fail(
            ErrorCode.NO_IMPORT, "Could not import rzpipe python library. Is it installed?"
        )

    if pwndbg.aglib.qemu.is_qemu_kernel():
        flags = ["-e", "bin.cache=true", "-e", "bin.relocs.apply=true"]
        if (kbase := pwndbg.aglib.kernel.kbase()) and filename == pwndbg.aglib.proc.exe():
            urand = pwndbg.aglib.elf.get_vmlinux_unrand_base(filename)
            if urand is not None:
                flags.extend(
                    [
                        "-e",
                        "bin.baddr=" + hex(kbase - urand),
                    ]
                )
        rz = rzpipe_mod.open(filename, flags)
    else:
        flags = ["-e", "io.cache=true"]
        exe = pwndbg.aglib.elf.exe()
        if pwndbg.aglib.elf.get_elf_info(filename).is_pie and exe is not None:
            flags.extend(["-B", hex(exe.address)])
        rz = rzpipe_mod.open(filename, flags=flags)
    # Lc -> list core plugins
    if "ghidra" not in str(rz.cmd("Lc")):
        return Status.coded_fail(
            ErrorCode.NO_IMPORT, "rizin plugin rzghidra must be installed and available from rz."
        )
    return rz


def rzcmd(arguments: list[str]) -> str:
    """
    Return result of rizin command or error string.
    """
    rz = rzpipe()
    if isinstance(rz, Status):
        # Since we got a status, it must be a failure.
        return message.error(rz.message)
    return str(rz.cmd(" ".join(arguments)))



from __future__ import annotations

from collections.abc import Mapping
from typing import NamedTuple


class Function(NamedTuple):
    type: str
    derefcnt: int
    name: str
    args: list[Argument]


class Argument(NamedTuple):
    type: str
    derefcnt: int
    name: str
    flags: tuple[Flag, ...] | None = None


class Flag(NamedTuple):
    value: int
    name: str


class LazyFunctions(Mapping[str, Function]):
    def __init__(self, *args, **kw):
        self._raw_dict = {}

    def __getitem__(self, key):
        if not self._raw_dict:
            from pwndbg.lib.functions_data import load_functions

            # dict is empty because functions have not been loaded yet
            self._raw_dict.update(load_functions())
        return self._raw_dict.__getitem__(key)

    def __iter__(self):
        return iter(self._raw_dict)

    def __len__(self):
        return len(self._raw_dict)


functions = LazyFunctions()


def format_flags_argument(flags: tuple[Flag, ...], value: int):
    original_value: int = value
    flag_names: list[str] = []

    # For some functions, some named flags are combinations
    # of other named flags. For example, the `mmap` flag
    # `MAP_SHARED_VALIDATE` is a combination of `MAP_SHARED`
    # and `MAP_PRIVATE`. As long as flags is ordered by
    # descending popcount, this loop will output more specific
    # flags.
    for flag in flags:
        if (value & flag.value) == flag.value:
            flag_names.append(flag.name)
            value = value & ~flag.value

    # If none of the known flags matched the value, just
    # format the value as a normal hex integer.
    if len(flag_names) == 0:
        return hex(original_value)

    # If we matched at least one known flag, but there
    # is some remaining un-matched portion of the value,
    # include that in the formatted | expression.
    if value != 0:
        flag_names.append(hex(value))

    # The final format includes the original value as hex,
    # any matched flags, and the left-over unmatched portion
    # of the integer.
    #
    # For example:
    # 0x03 (FLAG_2|0x01)
    return f"{original_value:#x} ({'|'.join(flag_names)})"



from __future__ import annotations


class IDTEntry:
    """
    Represents an entry in the Interrupt Descriptor Table (IDT)

    The IDTEntry class stores information about an IDT entry, including its index,
    offset, segment selector, descriptor privilege level (DPL), gate type, and
    interrupt stack table (IST) index.

    https://wiki.osdev.org/Interrupt_Descriptor_Table
    """

    def __init__(self, entry):
        self.offset = None
        self.segment = None
        self.dpl = None
        self.type = None
        self.ist = None
        self.present = None

        if len(entry) == 8:
            self._parse_entry32(entry)
        elif len(entry) == 16:
            self._parse_entry64(entry)

    def _parse_entry32(self, entry):
        """
        Parse a 32-bit IDT entry.

                                Gate Descriptor (32-bit)
        63                                  48  47  45 44        40               32
        +------------------------------------+--+---+--+---------+----------------+
        |                                    |P |DPL|0 |Gate Type|  Reserved      |
        |    Offset 31..16                   |  |   |  |         |                |
        |                                    |  |   |  |         |                |
        +------------------+------------------+------------------+----------------+
        31                                   16                                   0
        +-------------------------------------+------------------+----------------+
        |                                     |                                   |
        |          Segment Selector           |           Offset 15..0            |
        |                                     |                                   |
        +------------------+------------------+------------------+----------------+
        """
        entry = int.from_bytes(entry, byteorder="little")

        self.offset = entry & 0xFFFF
        self.offset |= ((entry >> 48) & 0xFFFF) << 16

        self.segment = (entry >> 16) & 0xFFFF
        self.type = (entry >> 40) & 0xF
        self.dpl = (entry >> 45) & 0x3
        self.present = (entry >> 47) & 0x1

    def _parse_entry64(self, entry):
        """Parse a 64-bit IDT entry."""
        entry = int.from_bytes(entry, byteorder="little")

        self.offset = entry & 0xFFFF
        self.offset |= ((entry >> 48) & 0xFFFF) << 16
        self.offset |= ((entry >> 64) & 0xFFFFFFFF) << 32

        self.segment = (entry >> 16) & 0xFFFF
        self.ist = (entry >> 32) & 0x7
        self.type = (entry >> 40) & 0xF
        self.dpl = (entry >> 45) & 0x3



from __future__ import annotations

from enum import Enum

import pwndbg
import pwndbg.aglib.kernel
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.color.message as message
import pwndbg.dbg_mod
import pwndbg.lib
import pwndbg.lib.cache
from pwndbg.aglib.kernel.macros import for_each_entry


class mod_mem_type(Enum):
    # Calculate runtime memory footprint by summing sizes of MOD_TEXT, MOD_DATA, MOD_RODATA, MOD_RO_AFTER_INIT,
    # which excludes initialization sections that are freed after the module load. See `enum mod_mem_type` in kernel source.
    MOD_TEXT = 0
    MOD_DATA = 1
    MOD_RODATA = 2
    MOD_RO_AFTER_INIT = 3  # might be empty
    # MOD_INIT_TEXT,
    # MOD_INIT_DATA,
    # MOD_INIT_RODATA,
    MOD_MEM_NUM_TYPES = 4


# TODO: handle potential negative offsets when CONFIG_RANDSTRUCT=y
@pwndbg.lib.cache.cache_until("stop")
def module_name_offset() -> int | None:
    modules = pwndbg.aglib.kernel.modules()
    if modules is None:
        print(message.warn("Could not find modules"))
        return None
    module = pwndbg.aglib.memory.read_pointer_width(int(modules))
    for i in range(0x100):
        offset = i * pwndbg.aglib.arch.ptrsize
        try:
            bs = pwndbg.aglib.memory.string(module + offset).decode("ascii")
            if len(bs) < 2:
                continue
            return offset
        except Exception:
            pass
    print(message.warn("Could not find module->name"))
    return None


@pwndbg.lib.cache.cache_until("stop")
def module_mem_offset() -> tuple[int | None, int | None, int | None]:
    modules = pwndbg.aglib.kernel.modules()
    if modules is None:
        print(message.warn("Could not find modules"))
        return None, None, None
    module = pwndbg.aglib.memory.read_pointer_width(int(modules))
    krelease = pwndbg.aglib.kernel.krelease()
    for i in range(0x100):
        offset = i * pwndbg.aglib.arch.ptrsize
        min_size = 0x10
        if krelease and (6, 13) <= krelease < (6, 15):
            min_size += 0x8
        for module_memory_size in (
            min_size,
            min_size + 0x38,
        ):
            found = True
            size_offset = None
            for mem_type in range(mod_mem_type.MOD_RO_AFTER_INIT.value):
                mem_ptr = module + offset + mem_type * module_memory_size
                if pwndbg.aglib.memory.peek(mem_ptr) is None:
                    found = False
                    break
                base = pwndbg.aglib.memory.read_pointer_width(mem_ptr)
                if base == 0 or ((base & 0xFFF) != 0):
                    found = False
                    break
                size_offset = pwndbg.aglib.arch.ptrsize
                if not krelease or (6, 15) <= krelease:
                    size_offset += 4
                elif (6, 13) <= krelease < (6, 15):
                    # https://elixir.bootlin.com/linux/v6.13/source/include/linux/module.h#L368
                    # additional fields were added
                    size_offset += pwndbg.aglib.arch.ptrsize + 4
                size = pwndbg.aglib.memory.u32(mem_ptr + size_offset)
                if not 0 < size < 0x100000:
                    found = False
                    break
            if found:
                return offset, module_memory_size, size_offset
    print(message.warn("Could not find module->mem"))
    return None, None, None


@pwndbg.lib.cache.cache_until("stop")
def module_layout_offset() -> tuple[int | None, int | None]:
    modules = pwndbg.aglib.kernel.modules()
    if modules is None:
        print(message.warn("Could not find modules"))
        return None, None
    module = pwndbg.aglib.memory.read_pointer_width(int(modules))
    for i in range(0x100):  # enough to search through the struct
        offset = i * pwndbg.aglib.arch.ptrsize
        ptr = module + offset + pwndbg.aglib.arch.ptrsize
        if pwndbg.aglib.memory.peek(ptr) is None:
            continue
        base = pwndbg.aglib.memory.read_pointer_width(ptr)
        if base == 0 or ((base & 0xFFF) != 0):
            continue
        valid = True
        for i in range(4):
            size = pwndbg.aglib.memory.u32(ptr + 4 * i)
            if not 0 < size < 0x100000:
                valid = False
                break
        if valid:
            return offset, offset + pwndbg.aglib.arch.ptrsize
    print(message.warn("Could not find module->init_layout"))
    return None, None


@pwndbg.lib.cache.cache_until("stop")
def module_kallsyms_offset() -> int | None:
    modules = pwndbg.aglib.kernel.modules()
    if modules is None:
        print(message.warn("Could not find modules"))
        return None
    module = pwndbg.aglib.memory.read_pointer_width(int(modules))
    krelease = pwndbg.aglib.kernel.krelease()
    for i in range(0x100):
        offset = i * pwndbg.aglib.arch.ptrsize
        ptr = module + offset
        if pwndbg.aglib.memory.peek(ptr) is None:
            continue
        kallsyms = pwndbg.aglib.memory.read_pointer_width(ptr)
        if pwndbg.aglib.memory.peek(kallsyms) is None or kallsyms == 0:
            continue
        symtab = pwndbg.aglib.memory.read_pointer_width(kallsyms)
        if pwndbg.aglib.memory.peek(symtab) is None:
            continue
        num_symtab = pwndbg.aglib.memory.read_pointer_width(kallsyms + pwndbg.aglib.arch.ptrsize)
        if pwndbg.aglib.memory.peek(num_symtab) is not None or num_symtab == 0:
            continue
        strtab = pwndbg.aglib.memory.read_pointer_width(kallsyms + pwndbg.aglib.arch.ptrsize * 2)
        if pwndbg.aglib.memory.peek(strtab) is None:
            continue
        if not krelease or krelease >= (5, 2):
            typetab = pwndbg.aglib.memory.read_pointer_width(
                kallsyms + pwndbg.aglib.arch.ptrsize * 3
            )
            if pwndbg.aglib.memory.peek(typetab) is None:
                continue
        return offset
    print(message.warn("Could not find module->kallsyms"))
    return None


@pwndbg.lib.cache.cache_until("stop")
def module_list_with_typeinfo() -> tuple[pwndbg.dbg_mod.Value, ...]:
    modules = pwndbg.aglib.kernel.modules()
    if modules is None:
        print(message.warn("Could not find modules"))
        return ()
    result = []
    head = pwndbg.aglib.memory.get_typed_pointer_value("struct list_head", modules)
    for module in for_each_entry(head, "struct module", "list"):
        result.append(module)
    # each entry if pointing to hte start of the module
    return tuple(result)


@pwndbg.lib.cache.cache_until("stop")
def module_list() -> tuple[int, ...]:
    modules = pwndbg.aglib.kernel.modules()
    if modules is None:
        print(message.warn("Could not find modules"))
        return ()
    modules = int(modules)
    result = []
    cur = pwndbg.aglib.memory.read_pointer_width(modules)
    while cur != modules:
        result.append(cur)
        cur = pwndbg.aglib.memory.read_pointer_width(cur)
    # each entry is pointing to the module->next
    return tuple(result)


def parse_module_kallsyms(kallsyms: int) -> list[tuple[str, int, str]]:
    is_64bit = pwndbg.aglib.arch.ptrsize == 8
    sizeof_symtab_entry = 24 if is_64bit else 16
    result = []
    symtab = pwndbg.aglib.memory.read_pointer_width(kallsyms)
    num_symtab = pwndbg.aglib.memory.read_pointer_width(kallsyms + pwndbg.aglib.arch.ptrsize)
    strtab = pwndbg.aglib.memory.read_pointer_width(kallsyms + pwndbg.aglib.arch.ptrsize * 2)
    typetab = 0
    krelease = pwndbg.aglib.kernel.krelease()
    if not krelease or krelease >= (5, 2):
        typetab = pwndbg.aglib.memory.read_pointer_width(kallsyms + pwndbg.aglib.arch.ptrsize * 3)
    strtab_offset = 0
    for i in range(num_symtab):
        sym_name = pwndbg.aglib.memory.string(strtab + strtab_offset).decode("utf-8")
        strtab_offset += len(sym_name) + 1
        if len(sym_name) == 0:
            continue
        sym_addr = pwndbg.aglib.memory.read_pointer_width(
            int(symtab) + sizeof_symtab_entry * i + pwndbg.aglib.arch.ptrsize
        )
        sym_type = None
        if not krelease or krelease >= (5, 2):
            sym_type = chr(pwndbg.aglib.memory.u8(typetab + i))
        else:
            sym_type = chr(
                pwndbg.aglib.memory.u8(symtab + sizeof_symtab_entry * i + 16 if is_64bit else 8)
            )
        result.append((sym_name, sym_addr, sym_type))
    return result


def all_modules_kallsyms() -> list[tuple[str, int, str]]:
    result = []
    if pwndbg.aglib.typeinfo.load("struct module") is not None:
        for module in module_list_with_typeinfo():
            if module.type.has_field("kallsyms"):
                kallsyms = int(module["kallsyms"])
                result += parse_module_kallsyms(kallsyms)
        return result
    offset = module_kallsyms_offset()
    if offset is not None:
        for module in module_list():
            kallsyms = pwndbg.aglib.memory.read_pointer_width(int(module) + offset)
            result += parse_module_kallsyms(kallsyms)
    return result



from __future__ import annotations

import argparse
import datetime
import os
import stat
import time
from typing import NamedTuple

import gdb
from tabulate import tabulate

import pwndbg.commands
import pwndbg.lib.config
from pwndbg.color import message
from pwndbg.commands import CommandCategory
from pwndbg.ui import get_window_size

_NONE = "none"
_OLDEST = "oldest"
_NEWEST = "newest"
_ASK = "ask"
_OPTIONS = [_NONE, _OLDEST, _NEWEST, _ASK]

pwndbg.config.add_param(
    "attachp-resolution-method",
    _ASK,
    "how to determine the process to attach when multiple candidates exists",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=_OPTIONS,
)

parser = argparse.ArgumentParser(
    description="""Attaches to a given pid, process name, process found with partial argv match or to a device file.

This command wraps the original GDB `attach` command to add the ability
to debug a process with a given name or partial name match. In such cases,
the process identifier is fetched via the `pidof <name>` command first. If no
matches are found, then it uses the `ps -eo pid,args` command to search for
partial name matches.

Original GDB attach command help:
    Attach to a process or file outside of GDB.
    This command attaches to another target, of the same type as your last
    "target" command ("info files" will show your target stack).
    The command may take as argument a process id or a device file.
    For a process id, you must have permission to send the process a signal,
    and it must have the same effective uid as the debugger.
    When using "attach" with a process id, the debugger finds the
    program running in the process, looking first in the current working
    directory, or (if not found there) using the source file search path
    (see the "directory" command).  You can also use the "file" command
    to specify the program, and to load its symbol table.""",
)


parser.add_argument("--no-truncate", action="store_true", help="dont truncate command args")
parser.add_argument("--retry", action="store_true", help="retry until a target is found")
parser.add_argument("--user", type=str, default=None, help="username or uid to filter by")
parser.add_argument(
    "-e",
    "--exact",
    action="store_true",
    help="get the pid only for an exact command name match",
)
parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="get pids also for partial cmdline matches etc",
)
parser.add_argument(
    "target",
    nargs="?",
    default=None,
    type=str,
    help="pid, process name, part of cmdline to be matched or device file to attach to (uses current loaded file name if not provided)",
)


import psutil


def find_pids(target: str, user: str | None, exact: bool, all: bool) -> list[int]:
    # Note: we can't use `ps -C <target>` because this does not accept process names with spaces
    # so target='a b' would actually match process names 'a' and 'b' here
    # so instead, we will filter by process name or full cmdline later on
    # if provided, filter by effective username or uid; otherwise, select all processes

    current_pid = os.getpid()

    pids_exact_match_cmd = []
    pids_partial_match_cmd = []
    pids_partial_match_args = []

    iter_process = psutil.process_iter(["pid", "name", "cmdline", "username"])
    iter_process = filter(lambda p: p.pid != current_pid, iter_process)
    if user is not None:
        iter_process = filter(lambda p: p.username() == user, iter_process)

    for proc in iter_process:
        try:
            cmdline = proc.cmdline()
        except (psutil.NoSuchProcess, psutil.AccessDenied):
            # The process no longer exists or we don't have permission
            continue
        else:
            if not cmdline:
                continue

            cmd = cmdline[0]
            args = " ".join(cmdline)

            if target == cmd:
                pids_exact_match_cmd.append(proc.pid)
            elif target in cmd:
                pids_partial_match_cmd.append(proc.pid)
            elif target in args:
                pids_partial_match_args.append(proc.pid)

    if exact and all:
        return pids_exact_match_cmd + pids_partial_match_cmd + pids_partial_match_args
    if exact:
        return pids_exact_match_cmd
    if all:
        return pids_exact_match_cmd + pids_partial_match_cmd + pids_partial_match_args
    return pids_exact_match_cmd or pids_partial_match_cmd or pids_partial_match_args


class ProcessInfo(NamedTuple):
    pid: int
    user: str
    elapsed: str
    command: str
    start_time: float


def resolve_target_process(
    pids: list[int],
    method: str,
    no_truncate: bool,
) -> int | None:
    proc_infos: list[ProcessInfo] = []

    for pid in pids:
        try:
            proc = psutil.Process(pid)
            start_time = proc.create_time()
            proc_infos.append(
                ProcessInfo(
                    proc.pid,
                    proc.username(),
                    str(datetime.timedelta(seconds=int(time.time() - start_time))),
                    " ".join(proc.cmdline()),
                    start_time,
                )
            )
        except (psutil.NoSuchProcess, psutil.AccessDenied, ValueError):
            continue

    if not proc_infos:
        print(message.error("No accessible processes found."))
        return None

    # Sort by start time (oldest first)
    proc_infos.sort(key=lambda x: x.start_time)

    if method == _OLDEST:
        return proc_infos[0].pid
    if method == _NEWEST:
        return proc_infos[-1].pid
    print(
        message.warn(
            f'Multiple processes found. Current resolution method is "{method}". '
            f"Run `config attachp-resolution-method` to see more information."
        )
    )

    headers = ["pid", "user", "elapsed", "command"]
    showindex: bool | range = False if method == _NONE else range(1, len(proc_infos) + 1)

    # Cast proc_infos for printing
    table_proc_infos = [
        [str(o.pid), str(o.user), str(o.elapsed), str(o.command)] for o in proc_infos
    ]

    # Calculate column width for truncation
    test_table = tabulate(table_proc_infos, headers=headers, showindex=showindex)
    table_width = len(test_table.splitlines()[1])
    max_command_width = max(len(row.command) for row in proc_infos)
    max_col_widths = max(max_command_width - (table_width - get_window_size()[1]), 10)

    # Truncate commands
    if not no_truncate:
        for row in table_proc_infos:
            row[3] = _truncate_string(row[3], max_col_widths)

    # Show the table
    msg = tabulate(
        table_proc_infos, headers=headers, showindex=showindex, maxcolwidths=max_col_widths
    )
    print(message.notice(msg))

    if method == _NONE:
        print(message.warn("Use `attach <pid>` to attach"))
        return None
    if method == _ASK:
        while True:
            prompt = message.notice(f"Which process to attach to? (1-{len(proc_infos)}) ")
            try:
                inp = input(prompt).strip()
            except EOFError:
                return None
            try:
                choice = int(inp)
                if 1 <= choice <= len(proc_infos):
                    return proc_infos[choice - 1].pid
            except ValueError:
                continue
    else:
        raise Exception("unreachable")


@pwndbg.commands.Command(parser, category=CommandCategory.START)
def attachp(
    target: str, no_truncate: bool, retry: bool, exact: bool, all: bool, user: str | None = None
) -> None:
    # As a default, the user may want to attach to a binary name taken from currently loaded file name
    if target is None:
        bin_path = pwndbg.aglib.proc.exe()
        if bin_path is None:
            print(
                message.error(
                    "No target name/pid/cmdline provided and no binary loaded in the debugger"
                )
            )
            print(message.error("(could not find the process name to attach to)"))
            return

        target = os.path.basename(bin_path)
    try:
        resolved_target = int(target)
    except ValueError:
        # GDB supposedly supports device files, so let's try it here...:
        #    <disconnect3d> hey, does anyone know what does `attach <device-file>` do?
        #    <disconnect3d> is this an alias for `target extended /dev/ttyACM0` or similar?
        #    <disconnect3d> I mean, `help attach` suggests that the `attach` command supports a device file target...
        #    <simark> I had no idea
        #    <simark> what you pass to attach is passed directly to target_ops::attach
        #    <simark> so it must be very target-specific
        #    <disconnect3d> how can it be target specific if it should  attach you to a target?
        #    <disconnect3d> or do you mean osabi/arch etc?
        #    <simark> in "attach foo", foo is interpreted by the target you are connected to
        #    <simark> But all targets I can find interpret foo as a PID
        #    <simark> So it might be that old targets had some other working mode
        if _is_device(target):
            resolved_target = target

        else:
            pids = find_pids(target, user, exact, all)
            if not pids and retry:
                user_filter = "" if not user else f" and user={user}"
                print(
                    message.warn(
                        f"Looking for pids for target={target}{user_filter} in a loop. Hit CTRL+C to cancel"
                    )
                )
                while not pids:
                    pids = find_pids(target, user, exact, all)

            if not pids:
                print(message.error(f"Process {target} not found"))
                return

            if len(pids) > 1:
                method = pwndbg.config.attachp_resolution_method
                resolved_target = resolve_target_process(pids, method.value, no_truncate)
            else:
                resolved_target = pids[0]

    if resolved_target is None:
        return

    print(message.on(f"Attaching to {resolved_target}"))
    try:
        gdb.execute(f"attach {resolved_target}")
    except gdb.error as e:
        print(message.error(f"Error: {e}"))
        return


def _is_device(path) -> bool:
    try:
        mode = os.stat(path).st_mode
    except FileNotFoundError:
        return False

    if stat.S_ISCHR(mode):
        return True

    return False


def _truncate_string(s: str, length: int):
    TRUNCATE_FILLER = " ... "
    if len(s) < length:
        return s
    truncate_point = (length - len(TRUNCATE_FILLER)) // 2
    result = s[:truncate_point]
    result += TRUNCATE_FILLER
    result += s[-(length - len(result)) :]
    return result



from __future__ import annotations

import pwndbg
import pwndbg.aglib.kernel.symbol
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo

#########################################
# structurs relevant to buddydump
#
#########################################
MAX_ORDER = 11


def find_zone_offsets() -> tuple[int, int, int, int, int]:
    pcp_off, name_off, freelist_off, pcp_pad, zone_sz = None, None, None, None, None
    node_data0 = pwndbg.aglib.kernel.node_data()
    if "CONFIG_NUMA" in pwndbg.aglib.kernel.kconfig():
        node_data0 = node_data0.dereference()
    node_data0 = int(node_data0)
    ptr = node_data0
    for i in range(20):  # the pcp offset should exist in those range
        val = pwndbg.aglib.memory.u64(ptr)
        ptr += 8
        if pwndbg.aglib.memory.is_kernel(val):
            # we have found `zone_pgdat`
            pcp_off = (i + 1) * 8
            break
    assert pcp_off, "can't find pcp offset"
    pcp_ptr = int(pwndbg.aglib.kernel.per_cpu(pwndbg.aglib.memory.u64(node_data0 + pcp_off)))
    for i in range(6):
        val = pwndbg.aglib.memory.u64(pcp_ptr + i * 8)
        if pwndbg.aglib.memory.is_kernel(val):
            pcp_pad = i * 8
            break
    assert pcp_pad, "can't find pcp pad"
    for i in range(20):
        char_ptr = pwndbg.aglib.memory.u64(ptr)
        ptr += 8
        name_str = None
        if pwndbg.aglib.memory.is_kernel(char_ptr):
            try:
                name_str = pwndbg.aglib.memory.string(char_ptr).decode()
            except Exception:
                continue
            if name_str is not None and name_str in pwndbg.aglib.kernel.symbol.POSSIBLE_ZONE_NAMES:
                name_off = i * 8 + pcp_off
                break
    assert name_off, "can't find name offset"
    prev = pwndbg.aglib.memory.u64(ptr)
    ptr += 8
    for i in range(1, 20):
        cur = pwndbg.aglib.memory.u64(ptr)
        ptr += 8
        if not pwndbg.aglib.memory.is_kernel(prev) and pwndbg.aglib.memory.is_kernel(cur):
            freelist_off = (i + 1) * 8 + name_off
            break
        prev = cur
    assert freelist_off, "can't find freelist offset"
    ptr += (
        MAX_ORDER * (pwndbg.aglib.kernel.symbol.nmtypes() * 0x10 + 8)
    ) + 0x10  # guessed MAX_ORDER * sizeof(struct list_head) + some other fields
    # find the next `zone_pgdat`
    for i in range(100):  # the pcp offset should exist in those range
        val = pwndbg.aglib.memory.u64(ptr)
        ptr += 8
        if pwndbg.aglib.memory.is_kernel(val):
            # we have found `zone_pgdat`
            zone_sz = ptr - pcp_off - node_data0
            break
    assert zone_sz and zone_sz < 0x4000 and zone_sz & 0xF == 0, (
        f"can't determine sizeof(struct zone) = {zone_sz}"
    )  # just to make sure it is sane
    return pcp_off, name_off, freelist_off, pcp_pad, zone_sz


@pwndbg.aglib.kernel.typeinfo_recovery("struct pglist_data", requires_kversion=True)
def recover_buddydump_typeinfo() -> str:
    nmtypes = pwndbg.aglib.kernel.symbol.nmtypes()
    nzones = pwndbg.aglib.kernel.symbol.nzones()
    nnodes = pwndbg.aglib.kernel.num_numa_nodes()
    npcplist = pwndbg.aglib.kernel.symbol.npcplist()

    result = f"#define KVERSION {pwndbg.aglib.kernel.symbol.kversion_cint()}\n"
    result += pwndbg.aglib.kernel.symbol.COMMON_TYPES
    pcp_off, name_off, freearea_off, pcp_pad, zone_sz = find_zone_offsets()
    if "CONFIG_NUMA" in pwndbg.aglib.kernel.kconfig():
        result += "#define CONFIG_NUMA\n"
    result += f"""
    struct free_area {{
        struct list_head	free_list[{nmtypes}];
        unsigned long		nr_free;
    }};
    """
    result += f"""
    struct per_cpu_pages {{
        char _pad[{pcp_pad}];
        /* Lists of pages, one per migrate type stored on the pcp-lists */
        struct list_head lists[{npcplist}]; // constant is sufficient for now
    }};
#if KVERSION < KERNEL_VERSION(5, 14, 0)
    struct per_cpu_pageset {{
        struct per_cpu_pages pcp;
    }};
#endif
/* custom type for page list data */
#ifdef CONFIG_NUMA
    typedef struct pglist_data *node_data_t[{nnodes}];
#else
    typedef struct pglist_data node_data_t;
#endif
    struct zone {{
        char _pad1[{pcp_off}];
#if KVERSION < KERNEL_VERSION(5, 14, 0)
        struct per_cpu_pageset *pageset;
#else
        struct per_cpu_pages *per_cpu_pageset;
#endif
        char _pad2[{name_off - pcp_off - 8}];
        char* name;
        char _pad3[{freearea_off - name_off - 8}];
        struct free_area free_area[{MAX_ORDER}]; // just defaults to 11 is sufficient here
        char _pad[{zone_sz - freearea_off - (MAX_ORDER * (nmtypes * 0x10 + 8))}];
    }};
    """
    result += f"""
    typedef struct pglist_data {{
        struct zone node_zones[{nzones}];
        // ... the rest of the fields are not important
    }} pg_data_t;
    """
    return result



from __future__ import annotations

import argparse

import pwndbg
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.color as color
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""Decode X86-64 GDT entries at address

See also:

* https://wiki.osdev.org/Global_Descriptor_Table
* https://wiki.osdev.org/GDT_Tutorial
""",
)

parser.add_argument(
    "address",
    type=int,
    help="x86-64 GDTR base address (e.g. read from sgdt instruction from [16:79] bits)",
)

parser.add_argument(
    "count",
    nargs="?",
    default=8,
    help="Number of entries to dump (should be (GDTR.size+1)/8)",
    type=int,
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MEMORY,
    notes="""
In 64-bit mode, the Base and Limit values are ignored, each descriptor covers
the entire linear address space regardless of what they are set to.
""",
)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.aglib.proc.OnlyWithArch(["x86-64"])
def gdt(address, count) -> None:
    address = int(address)
    count = int(count)
    idx = 0

    print("Dumping GDT entries:")

    while idx < count:
        addr = address + idx * 8
        value = pwndbg.aglib.memory.u64(addr)
        e = decode_gdt_entry(value)

        print(f"{addr:#x} => {e}")
        idx += 1


def decode_gdt_entry(value):
    if value == 0:
        return "<NULL entry>"

    limit_low = value & 0xFFFF
    limit_high = (value >> 48) & 0xF
    limit = (limit_high << 16) | limit_low

    base_low = (value >> 16) & 0xFFFF
    base_mid = (value >> 32) & 0xFF
    base_high = (value >> 56) & 0xFF
    base = (base_high << 24) | (base_mid << 16) | base_low

    access_byte = (value >> 40) & 0xFF

    flags = (value >> 52) & 0xF

    # If GDT descriptor is stored in RO pages and this bit is 0
    # the CPU trying to set this bit will trigger page fault
    accessed_bit = access_byte & (1 << 0)

    # If 0, read (code segment) or write (data segment) access is not allowed
    # (Read is always allowed for data; write is never allowed for code segment)
    rw_bit = (access_byte & (1 << 1)) >> 1

    # Direction bit/Conforming bit
    dc_bit = (access_byte & (1 << 2)) >> 2

    # If 0 == data segment, if 1 == code segment
    exec_bit = (access_byte & (1 << 3)) >> 3

    # Descriptor type; 0 == system segment, 1 == code or data segment
    type_bit = (access_byte & (1 << 4)) >> 4

    # Descriptor privilege level; 0 == kernel, 3 == userspace
    dpl_bits = (access_byte & (0b11 << 5)) >> 5

    # Must be set for a valid segment
    present_bit = (access_byte & (1 << 7)) >> 7

    colorme = lambda label, val: (color.green if val else color.red)(label)

    access_str = "|".join(
        (
            colorme("P", present_bit),
            f"DPL:{dpl_bits}",
            colorme("S", type_bit),
            colorme("E", exec_bit),
            colorme("DC", dc_bit),
            colorme("RW", rw_bit),
            colorme("A", accessed_bit),
        )
    )

    flags_str = "|".join(
        (
            colorme("G", (flags & (1 << 3)) >> 3),
            colorme("DB", (flags & (1 << 2)) >> 2),
            colorme("L", (flags & (1 << 1)) >> 1),
        )
    )

    return f"base={base:#4x}, limit={limit:#8x}, access={access_str}, flags={flags_str}"



"""
Specific command output wrappers.
"""

from __future__ import annotations

import functools
import subprocess
from collections.abc import Callable
from subprocess import STDOUT
from typing import TypeVar

from pwnlib.util.misc import which
from typing_extensions import ParamSpec

import pwndbg.commands

P = ParamSpec("P")
T = TypeVar("T")


class OnlyWithCommand:
    def __init__(self, *commands: str | list[str]) -> None:
        self.all_cmds: list[str] = [cmd[0] if isinstance(cmd, list) else cmd for cmd in commands]
        for command in commands:
            self.cmd: list[str] = command if isinstance(command, list) else [command]
            self.cmd_path: str | None = which(self.cmd[0])
            if self.cmd_path:
                break

    def __call__(self, function: Callable[P, T]) -> Callable[P, T | None]:
        function.cmd = self.cmd

        @pwndbg.commands.OnlyWithFile
        @functools.wraps(function)
        def _OnlyWithCommand(*a: P.args, **kw: P.kwargs) -> T | None:
            if self.cmd_path:
                return function(*a, **kw)
            raise OSError(f"Could not find command(s) {', '.join(self.all_cmds)} in $PATH")

        return _OnlyWithCommand


def call_cmd(*cmd: str) -> str:
    return subprocess.check_output(cmd, stderr=STDOUT).decode("utf-8")



from __future__ import annotations

import argparse

import pwndbg.aglib.godbg
import pwndbg.commands
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="Dumps a Go value of a given type at a specified address."
)
parser.add_argument(
    "ty",
    type=str,
    help="Go type of value to dump, e.g. map[int]string, or the address of a type to resolve at runtime, e.g. 0x408860",
)
parser.add_argument(
    "address",
    type=pwndbg.commands.AddressExpr,
    help="Address to dump",
)
parser.add_argument("-x", "--hex", action="store_true", help="Display non-pointer integers as hex")
parser.add_argument(
    "-f",
    "--decimals",
    nargs="?",
    type=int,
    help="Configures the number of decimal places to display for floating points",
)

parser.add_argument(
    "-d",
    "--debug",
    action="store_true",
    help="Shows debug info, like addresses for slice/map elements, slice capacity, etc.",
)

parser.add_argument(
    "-p",
    "--pretty",
    action="store_true",
    help="Enables pretty printing",
)


@pwndbg.commands.Command(
    parser, category=CommandCategory.MEMORY, command_name="go-dump", aliases=["god"]
)
@pwndbg.commands.OnlyWhenRunning
def go_dump(
    ty: str, address: int, hex: bool, decimals: int | None, debug: bool, pretty: bool
) -> None:
    try:
        ty_addr = int(ty, 0)
        (_, parsed_ty) = pwndbg.aglib.godbg.decode_runtime_type(ty_addr)
        if parsed_ty is None:
            print("Failed to decode runtime type.")
            return
    except ValueError:
        parsed_ty = pwndbg.aglib.godbg.parse_type(ty)
    fmt = pwndbg.aglib.godbg.FormatOpts(
        int_hex=hex, float_decimals=decimals, debug=debug, pretty=pretty
    )
    print(parsed_ty.dump(address, fmt))


parser = argparse.ArgumentParser(
    description="Dumps a Go runtime reflection type at a specified address."
)
parser.add_argument(
    "address",
    type=pwndbg.commands.AddressExpr,
    help="Address to dump",
)


@pwndbg.commands.Command(
    parser, category=CommandCategory.MEMORY, command_name="go-type", aliases=["goty"]
)
@pwndbg.commands.OnlyWhenRunning
def go_type(address: int) -> None:
    meta, ty = pwndbg.aglib.godbg.decode_runtime_type(address, True)
    print(f" Name: {meta.name}")
    print(f" Kind: {meta.kind.name}")
    print(f" Size: {meta.size} ({meta.size:#x})")
    print(f"Align: {meta.align}")
    print(f"Parse: {ty}")
    if ty:
        data = ty.additional_metadata()
        if data:
            print("\n".join(data))



"""
Implements handling of musl's allocator mallocng.
https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng
"""

from __future__ import annotations

from enum import Enum

from typing_extensions import override

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.heap.heap
import pwndbg.aglib.memory as memory
import pwndbg.aglib.stack
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.auxv
import pwndbg.color.message as message
import pwndbg.dbg_mod
import pwndbg.search

# https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L14
# Slot granularity.
UNIT: int = 16
# Size of in-band metadata.
IB: int = 4

# https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/malloc.c#L12
# Describes the possible sizes a slot can be. These are `/ UNIT`.
# fmt: off
size_classes: list[int] = [
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 15, 18, 20,
    25, 31, 36, 42, 50, 63, 72, 84, 102, 127, 146,
    170, 204, 255, 292, 340, 409, 511, 584, 682, 818,
    1023, 1169, 1364, 1637, 2047, 2340, 2730, 3276,
    4095, 4680, 5460, 6552, 8191,
]
# fmt: on


class SlotState(Enum):
    ALLOCATED = "allocated"
    FREED = "freed"
    # Available - this slot has not yet been allocated.
    AVAIL = "available"


# Shorthand
def int_size() -> int:
    return pwndbg.aglib.typeinfo.sint.sizeof


class Group:
    """
    A group is an array of slots.

    https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L17
    struct group {
      struct meta *meta;
      unsigned char active_idx:5;
      char pad[UNIT - sizeof(struct meta *) - 1];
      unsigned char storage[];
    };
    """

    def __init__(self, addr: int) -> None:
        self.addr = addr

        self._meta = None
        self._active_idx = None

    def preload(self) -> None:
        """
        Read all the necessary process memory to populate the group's
        fields.

        Do this if you know you will be using most of the
        fields of the group. It will be faster, since we can do one
        reads instead of two small ones. You may also catch
        inaccessible memory exceptions here and not worry about it later.

        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        data = memory.read(self.addr, pwndbg.aglib.arch.ptrsize + 1)
        self._meta = Meta(pwndbg.aglib.arch.unpack(data[: pwndbg.aglib.arch.ptrsize]))
        self._active_idx = data[-1] & 0b11111

    @property
    def meta(self) -> Meta:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._meta is None:
            self._meta = Meta(memory.read_pointer_width(self.addr))

        return self._meta

    @property
    def active_idx(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._active_idx is None:
            self._active_idx = memory.u8(self.addr + pwndbg.aglib.arch.ptrsize) & 0b11111

        return self._active_idx

    @property
    def storage(self) -> int:
        return self.addr + UNIT

    @property
    def group_size(self) -> int:
        """
        The size of this group, in bytes.

        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/malloc.c#L234
        return self.meta.stride * self.meta.cnt + UNIT

    def set_meta(self, meta: Meta) -> None:
        """
        Sets the meta object for this group.

        If the meta for this group is already calculated by the callee,
        use this to prevent it from being wastefully recalculated.
        """
        self._meta = meta

    def at_index(self, idx: int) -> int:
        """
        Get the address of the slot at index idx.
        """
        return self.storage + idx * self.meta.stride


class Slot:
    """
    The "unit of allocation" (analogous to glibc's "chunk").
    There is no struct in the source code that describes it.
    """

    def __init__(self, p: int) -> None:
        # The start of user memory. It may
        # not be the actual start of the slot.
        self.p: int = p

        # == The p header fields.
        self._offset: int | None = None
        # p[-3]. Stores lot's of different kinds of
        # information.
        self._pn3: int | None = None
        self._idx: int | None = None
        self._reserved_hd: int | None = None
        self._big_offset_check: int | None = None
        # ==

        # == The footer fields.
        self._reserved_ft: int | None = None
        # ==

        # == The start header fields.
        self._start: int | None = None
        self._cyclic_offset: int | None = None
        # start[-3]. Stores whether we are cyclic.
        self._startn3: int | None = None
        # ==

        self._reserved: int | None = None
        self._group: Group | None = None
        self._meta: Meta | None = None
        self._slot_state: SlotState | None = None

    def preload(self) -> None:
        """
        Read all the necessary process memory to populate the slot's
        p header fields.

        Do this if you know you will be using most of the
        fields of the slot. It will be faster, since we can do a few
        big reads instead of many small ones. You may also catch
        inaccessible memory exceptions here and not worry about it later.

        Fields dependant on the meta are not loaded - you will still
        need to worry about exceptions coming from them.

        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # == Read the p header.
        pheader = memory.read(self.p - 8, 8)

        self._big_offset_check = pheader[4]
        if self._big_offset_check:
            self._offset = int.from_bytes(pheader[0:4], pwndbg.aglib.arch.endian, signed=False)
        else:
            self._offset = int.from_bytes(pheader[6:8], pwndbg.aglib.arch.endian, signed=False)
        self._pn3 = pheader[5]
        # ==

        # To calculate footer and start header fields
        # we need self.meta.stride. However we want to be able to
        # return some information even if the meta is corrupt or
        # unreachable (e.g. this slot is freed or avail), so
        # we won't load that here.

        # Other fields are calculated without memory reads.

    def preload_meta_dependants(self) -> None:
        """
        Preloads all fields that depend on a sane meta.

        It generally only makes sense to run this after preload().
        Calling this reduces the amount of process writes and centralizes
        field exceptions to this function.

        If both preload() and preload_meta_dependants() return without
        exceptions, all the fields in this class are guaranteed to not
        cause any more memory reads nor raise any more exceptions.

        Raises:
            pwndbg.dbg_mod.Error: When the meta is corrupt and/or
                reading memory fails.
        """
        # Make sure stride is valid.
        _ = self.meta.stride

        # Read the start header only if we need to.
        if self.start != self.p:
            startheader = memory.read(self.start - 3, 3)
            self._startn3 = int.from_bytes(startheader[0:1], pwndbg.aglib.arch.endian, signed=False)
            self._cyclic_offset = int.from_bytes(
                startheader[1:3], pwndbg.aglib.arch.endian, signed=False
            )

        # Read footer.
        if self.reserved_in_header != 5:
            self._reserved_ft = -1
        else:
            self._reserved_ft = memory.u32(self.end - 4)

        # Other fields are calculated without memory reads.

    # p header fields..

    @property
    def offset(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L132
        if self._offset is None:
            if self.big_offset_check:
                # This can only happen in aligned allocations, which is kind of
                # weird. All allocations of this size are probably mmaped.
                # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/aligned_alloc.c#L49
                self._offest = memory.u32(self.p - 8)
            else:
                self._offset = memory.u16(self.p - 2)

        return self._offset

    @property
    def pn3(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._pn3 is None:
            self._pn3 = memory.u8(self.p - 3)

        return self._pn3

    @property
    def idx(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L133
        if self._idx is None:
            if self.pn3 == 255:
                # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/donate.c#L29
                self._idx = 0
            else:
                self._idx = self.pn3 & 31

        return self._idx

    @property
    def reserved_in_header(self) -> int:
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L193
        if self._reserved_hd is None:
            self._reserved_hd = self.pn3 >> 5

        return self._reserved_hd

    @property
    def big_offset_check(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L134
        if self._big_offset_check is None:
            self._big_offset_check = memory.u8(self.p - 4)

        return self._big_offset_check

    # start header fields..

    @property
    def start(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        # We have this if-statement so Slot.from_start() can
        # populate _start, giving us lots of fields even with
        # a corrupt meta.
        if self._start is None:
            # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/free.c#L108
            self._start = self.group.storage + self.meta.stride * self.idx

        return self._start

    @property
    def cyclic_offset(self) -> int:
        """
        Returns zero if is_cyclic() is False.

        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L216
        # Not sure why musl saves it, it doesn't seem to use it.
        # We could calculate it more easily than musl does `(self.p - self.start) // UNIT`
        # but let's report the actual in-band metadata in case the structure
        # is partially corrupted.
        if self._cyclic_offset is None:
            if self.is_cyclic():
                self._cyclic_offset = memory.u16(self.start - 2)
            else:
                self._cyclic_offset = 0

        return self._cyclic_offset

    @property
    def startn3(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._startn3 is None:
            if self.p == self.start:
                # No need to read memory twice.
                self._startn3 = self.pn3
            else:
                self._startn3 = memory.u8(self.start - 3)

        return self._startn3

    # footer fields..

    @property
    def reserved_in_footer(self) -> int:
        """
        Returns -1 if the value is invalid, i.e.
        reserved_in_header() != 5.

        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L161
        if self._reserved_ft is None:
            if self.reserved_in_header != 5:
                self._reserved_ft = -1
            else:
                self._reserved_ft = memory.u32(self.end - 4)

        return self._reserved_ft

    # code variables..

    @property
    def end(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/free.c#L109
        return self.start + self.meta.stride - IB

    @property
    def reserved(self) -> int:
        """
        Returns 0 if reserved_in_header() == 6.
        Returns -1 if reserved_in_header() == 7.

        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L161
        # Lots of asserts here..
        if self._reserved is None:
            if self.reserved_in_header < 5:
                self._reserved = self.reserved_in_header
            elif self.reserved_in_header == 5:
                self._reserved = self.reserved_in_footer
            elif self.reserved_in_header == 6:
                # See contains_group()
                self._reserved = 0
            else:
                # Value forced due to bit-size.
                assert self.reserved_in_header == 7
                # It is possible for start[-3] to contain (7<<5),
                # but p[-3] shouldn't unless the slot is free.
                self._reserved = -1

        return self._reserved

    @property
    def nominal_size(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        # Special case (probably) freed slots (see Slot.reserved):
        if self.reserved == -1:
            # Returning the value calculated below would confuse users
            # as, semantically, the slot has no size.
            return 0

        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L159
        return self.end - self.reserved - self.p

    @property
    def user_size(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        return self.nominal_size

    @property
    def slack(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading meta fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L199
        return (self.meta.stride - self.nominal_size - IB) // UNIT

    # custom..

    @property
    def group(self) -> Group:
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L139
        if self._group is None:
            self._group = Group(self.p - UNIT * self.offset - UNIT)

        return self._group

    @property
    def meta(self) -> Meta:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L140
        if self._meta is None:
            self._meta = Meta(memory.read_pointer_width(self.group.addr))

        return self._meta

    @property
    def slot_state(self) -> SlotState:
        if self._slot_state is None:
            # The actual "source of truth" for slot allocation state is
            # self.meta.slotstate_at_index() however we can only resolve
            # the meta if the state is ALLOCATED.
            # We will do a heuristic check that should be good in most cases.

            meta_says: SlotState | None = None
            try:
                meta_says = self.meta.slotstate_at_index(self.idx)
            except pwndbg.dbg_mod.Error:
                # We can't reach the meta. Either the slot is not allocated
                # or it is allocated but the meta pointer is corrupted.
                meta_says = None

            if meta_says is not None:
                self._slot_state = meta_says
            else:
                # When a slot is freed, its p[-3] gets set to 0xFF so the
                # offset to group start (and by extension, meta) is unrecoverable.
                # We will check for this, although musl only ever sets this
                # and never uses this as a source of truth.
                if self.pn3 == 0xFF:
                    # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/free.c#L112
                    self._slot_state = SlotState.FREED
                else:
                    self._slot_state = SlotState.AVAIL

        return self._slot_state

    # checks..

    def is_cyclic(self) -> int:
        """
        Returns whether mallocng reports that p != start.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L217
        # We could of course just do `return p != start`
        # but we want to report the actual metadata in case the structure
        # is partially corrupted.
        return self.startn3 == 224

    def contains_group(self) -> bool:
        """
        Does this slot nest a group?
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/malloc.c#L269
        return self.reserved_in_header == 6

    # external setters..

    def set_group(self, group: Group) -> None:
        """
        If the slot is FREED or AVAIL, it is impossible for it to
        recover the start of its group, and ergo its meta.

        You can thus use this to set it externally.
        """
        self._group = group

    # constructors..

    @classmethod
    def from_p(cls, p: int) -> Slot:
        return cls(p)

    @classmethod
    def from_start(cls, start: int) -> Slot:
        # We need to check if we are cyclic or not.
        # See is_cyclic() and cyclic_offset() logic.
        sn3 = memory.u8(start - 3)
        if sn3 == 224:
            off = memory.u16(start - 2)
            obj = cls(start + off * UNIT)
            obj._startn3 = sn3
        else:
            # freed / avail slots will also go into this branch.
            obj = cls(start)
            obj._startn3 = obj._pn3 = sn3

        obj._start = start

        return obj


class GroupedSlot:
    """
    This is *not* a mallocng concept, this is a pwndbg abstraction.

    A Slot object uses its inband metadata to recover all its fields and
    uncover more information about itself by locating its group and meta.
    It works essentially the same way mallocng's free() works.

    However, if a slot is freed or available, most of its in-band metadata
    will be invalid and it will not be able to recover group and meta. But,
    given the start of the slot, we can infer which group it belongs to and
    what its index is by walking allocator state i.e. ctx i.e. by using
    Mallocng.find_slot().

    A GroupedSlot then describes all information we can glean about a slot
    which is described by a (group, idx) pair. Many of its fields can be
    completely different from a Slot at the same location. They are guaranteed
    to be the same only if the slot is ALLOCATED and hasn't been corrupted.

    Not all fields that are available in Slot are available in GroupedSlot.

    Make sure the group you are passing to the constructor points to a valid meta
    object.
    """

    def __init__(self, group: Group, idx: int) -> None:
        self.group = group
        self.meta = self.group.meta
        self.idx = idx
        self.stride = self.meta.stride
        self.slot_state = self.meta.slotstate_at_index(self.idx)
        self.start = self.group.storage + self.meta.stride * self.idx
        self.end = self.start + self.stride - IB


class Meta:
    """
    The metadata of a group.

    https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L24
    struct meta {
      struct meta *prev, *next;
      struct group *mem;
      volatile int avail_mask, freed_mask;
      uintptr_t last_idx:5;
      uintptr_t freeable:1;
      uintptr_t sizeclass:6;
      uintptr_t maplen:8*sizeof(uintptr_t)-12;
    };
    """

    def __init__(self, addr: int) -> None:
        self.addr: int = addr

        self._prev: int | None = None
        self._next: int | None = None
        self._mem: int | None = None
        self._avail_mask: int | None = None
        self._freed_mask: int | None = None
        self._last_idx: int | None = None
        self._freeable: int | None = None
        self._sizeclass: int | None = None
        self._maplen: int | None = None

        self._stride: int | None = None

    def preload(self) -> None:
        """
        Read all the necessary process memory to populate the meta's
        fields.

        Do this if you know you will be using most of the
        fields of the meta. It will be faster, since we can do a one
        big read instead of many small ones. You may also catch
        inaccessible memory exceptions here and not worry about it later.

        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        ptrsize = pwndbg.aglib.arch.ptrsize
        endian = pwndbg.aglib.arch.endian

        # Read the whole struct.
        data = memory.read(self.addr, Meta.sizeof())

        cur_offset = 0

        def next_int(size: int, signed: bool = False) -> int:
            nonlocal cur_offset
            val = int.from_bytes(data[cur_offset : (cur_offset + size)], endian, signed=signed)
            cur_offset += size
            return val

        self._prev = next_int(ptrsize)
        self._next = next_int(ptrsize)
        self._mem = next_int(ptrsize)
        self._avail_mask = next_int(int_size())
        self._freed_mask = next_int(int_size())
        # I think this is how I should read a bitfield.
        # http://mjfrazer.org/mjfrazer/bitfields/
        flags = next_int(ptrsize)
        self._last_idx = flags & 0b11111
        self._freeable = (flags >> 5) & 1
        self._sizeclass = (flags >> 6) & 0b111111
        self._maplen = flags >> 12

        assert cur_offset == Meta.sizeof()

        # All other values are calculated without
        # memory reads.

    @property
    def prev(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._prev is None:
            self._prev = memory.read_pointer_width(self.addr)

        return self._prev

    @property
    def next(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._next is None:
            self._next = memory.read_pointer_width(self.addr + pwndbg.aglib.arch.ptrsize)

        return self._next

    @property
    def mem(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._mem is None:
            self._mem = memory.read_pointer_width(self.addr + pwndbg.aglib.arch.ptrsize * 2)

        return self._mem

    @property
    def avail_mask(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._avail_mask is None:
            # While the type is technically a signed int, it makes more
            # sense to interpret it as unsigned semantically.
            self._avail_mask = memory.uint(self.addr + pwndbg.aglib.arch.ptrsize * 3)

        return self._avail_mask

    @property
    def freed_mask(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._freed_mask is None:
            offset = pwndbg.aglib.arch.ptrsize * 3 + int_size()
            # Technically signed.
            self._freed_mask = memory.uint(self.addr + offset)

        return self._freed_mask

    @property
    def last_idx(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._last_idx is None:
            offset = pwndbg.aglib.arch.ptrsize * 3 + int_size() * 2
            # reading pointer width so it works regardless of endianness
            self._last_idx = memory.read_pointer_width(self.addr + offset) & 0b11111

        return self._last_idx

    @property
    def freeable(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._freeable is None:
            offset = pwndbg.aglib.arch.ptrsize * 3 + int_size() * 2
            self._freeable = (memory.read_pointer_width(self.addr + offset) >> 5) & 1

        return self._freeable

    @property
    def sizeclass(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._sizeclass is None:
            offset = pwndbg.aglib.arch.ptrsize * 3 + int_size() * 2
            self._sizeclass = (memory.read_pointer_width(self.addr + offset) >> 6) & 0b111111

        return self._sizeclass

    @property
    def maplen(self) -> int:
        """
        Raises:
            pwndbg.dbg_mod.Error: When reading memory fails.
        """
        if self._maplen is None:
            offset = pwndbg.aglib.arch.ptrsize * 3 + int_size() * 2
            self._maplen = memory.read_pointer_width(self.addr + offset) >> 12

        return self._maplen

    # Semi-custom methods..

    @property
    def stride(self) -> int:
        """
        Returns -1 if sizeclass >= len(size_classes).
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L175
        if self._stride is None:
            if not self.last_idx and self.maplen:
                self._stride = self.maplen * 4096 - UNIT
            else:
                if self.sizeclass < len(size_classes):
                    self._stride = UNIT * size_classes[self.sizeclass]
                else:
                    # The meta is corrupted.
                    self._stride = -1

        return self._stride

    # Custom methods..

    @property
    def cnt(self) -> int:
        """
        Number of slots in the group.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/free.c#L60
        return self.last_idx + 1

    @property
    def is_donated(self) -> bool:
        """
        Returns whether the group object referred to by this meta has been
        created by being donated by ld.
        """
        # When mapped object files contain unused memory, they are donated
        # to the heap. See https://elixir.bootlin.com/musl/v1.2.5/source/ldso/dynlink.c#L600
        # and https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/donate.c#L36 .
        # Only in this case is `meta.freeable = 0;`
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/donate.c#L25
        return not self.freeable

    @property
    def is_mmaped(self) -> bool:
        """
        Returns whether the group object referred to by this meta has been
        created by being mmaped.
        """
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L177
        # The if-statement in the source also checks for !g->last_idx but in practice
        # I've seen this value be non-zero for mmap()-ed groups, so we're leaving it out.
        return not self.is_donated and bool(self.maplen)

    @property
    def is_nested(self) -> bool:
        """
        Returns whether the group object referred to by this meta has been
        created by being nested into a slot.
        """
        return not self.is_donated and not self.is_mmaped

    def parent_group(self) -> int:
        """
        If this group is nested, returns the address of the group which
        contains the slot in which this group is in. Otherwise, returns -1.
        """
        if not self.is_nested:
            return -1

        return Slot(Group(self.mem).addr).group.addr

    def root_group(self) -> Group:
        """
        Returns the topmost/biggest parent group. It will never be a nested
        group. If this group isn't nested, this group is returned.
        """
        cur: Group = Group(self.mem)

        while cur.meta.is_nested:
            cur = Slot(cur.addr).group

        return cur

    def slotstate_at_index(self, idx: int) -> SlotState:
        me = 1 << idx
        if self.freed_mask & me:
            return SlotState.FREED
        if self.avail_mask & me:
            return SlotState.AVAIL
        return SlotState.ALLOCATED

    @staticmethod
    def sizeof() -> int:
        return 2 * int_size() + 4 * pwndbg.aglib.arch.ptrsize


class MetaArea:
    """
    Slabs that contain metas, linked in a singly-linked list.

    https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L34
    struct meta_area {
      uint64_t check;
      struct meta_area *next;
      int nslots;
      struct meta slots[];
    };
    """

    def __init__(self, addr: int) -> None:
        self.addr: int = addr

        self.check: int = 0
        self.meta_area: int = 0
        self.nslots: int = 0
        self.slots: int = 0

        self.load()

    def load(self) -> None:
        ptrsize = pwndbg.aglib.arch.ptrsize
        uint64size = pwndbg.aglib.typeinfo.uint64.sizeof
        endian = pwndbg.aglib.arch.endian

        data: bytearray = memory.read(self.addr, uint64size + ptrsize + int_size())

        cur_offset = 0

        def next_int(size: int, signed: bool = False) -> int:
            nonlocal cur_offset
            val = int.from_bytes(data[cur_offset : (cur_offset + size)], endian, signed=signed)
            cur_offset += size
            return val

        self.check = next_int(uint64size)
        self.next = next_int(ptrsize)
        self.nslots = next_int(int_size(), True)

        # Alignment adjustment
        cur_offset += ptrsize - int_size()

        self.slots = self.addr + cur_offset

    def at_index(self, idx: int) -> int:
        """
        Returns the address of the meta object located
        at index idx.
        """
        return self.slots + idx * Meta.sizeof()

    @property
    def area_size(self) -> int:
        """
        Returns not the size of `struct meta_area` but rather
        the size of the memory this object represents.
        """
        return (self.slots - self.addr) + self.nslots * Meta.sizeof()


class MallocContext:
    """
    The global object that holds all allocator state.

    https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L41
    struct malloc_context {
      uint64_t secret;
    #ifndef PAGESIZE
      size_t pagesize;
    #endif
      int init_done;
      unsigned mmap_counter;
      struct meta *free_meta_head;
      struct meta *avail_meta;
      size_t avail_meta_count, avail_meta_area_count, meta_alloc_shift;
      struct meta_area *meta_area_head, *meta_area_tail;
      unsigned char *avail_meta_areas;
      struct meta *active[48];
      size_t usage_by_class[48];
      uint8_t unmap_seq[32], bounces[32];
      uint8_t seq;
      uintptr_t brk;
    };
    """

    def __init__(self, addr: int) -> None:
        self.addr: int = addr

        self.secret: int = 0
        self.pagesize: int = 0
        self.init_done: int = 0
        self.mmap_counter: int = 0
        self.free_meta_head: int = 0
        self.avail_meta: int = 0
        self.avail_meta_count: int = 0
        self.avail_meta_area_count: int = 0
        self.meta_alloc_shift: int = 0
        self.meta_area_head: int = 0
        self.meta_area_tail: int = 0
        self.avail_meta_areas: int = 0
        self.active: list[int] = []
        self.usage_by_class: list[int] = []
        self.unmap_seq: list[int] = []
        self.bounces: list[int] = []
        self.seq: int = 0
        self.brk: int = 0

        self.sizeof: int = 0
        self.has_pagesize_field: bool = False

        self.load()

    def load(self) -> None:
        ptrsize = pwndbg.aglib.arch.ptrsize
        size_tsize = pwndbg.aglib.typeinfo.size_t.sizeof
        unsignedsize = pwndbg.aglib.typeinfo.uint.sizeof
        uint8size = pwndbg.aglib.typeinfo.uint8.sizeof
        uint64size = pwndbg.aglib.typeinfo.uint64.sizeof
        endian = pwndbg.aglib.arch.endian

        # We will assume the struct has the pagesize field at first (even though it usually
        # doesn't), which allows us to only do one memory read. This is 0x3A8 bytes on x86_64.
        self.sizeof = uint64size + size_tsize + int_size() + unsignedsize + ptrsize * 2
        self.sizeof += size_tsize * 3 + ptrsize * 2 + ptrsize + ptrsize * 48 + size_tsize * 48
        self.sizeof += uint8size * 32 * 2 + uint8size + (ptrsize - uint8size) + ptrsize

        data: bytearray = memory.read(self.addr, self.sizeof)

        cur_offset = 0

        def next_int(size: int, signed: bool = False) -> int:
            nonlocal cur_offset
            val = int.from_bytes(data[cur_offset : (cur_offset + size)], endian, signed=signed)
            cur_offset += size
            return val

        self.secret = next_int(uint64size)

        # We will read `int` bytes past the `secret`. The `init_done` field can only contain
        # values 0 and 1, so if we get that we know the struct doesn't have the pagesize field.
        # If it contains a value > 1 it must be describing a page size.
        something = int.from_bytes(
            data[cur_offset : (cur_offset + int_size())], endian, signed=True
        )
        self.has_pagesize_field = something > 1

        if self.has_pagesize_field:
            self.pagesize = next_int(size_tsize)
            self.init_done = next_int(int_size(), True)
        else:
            self.init_done = something
            cur_offset += int_size()

            # Fix our assumption, we don't have `size_t pagesize` field.
            self.sizeof -= size_tsize

        self.mmap_counter = next_int(unsignedsize)
        self.free_meta_head = next_int(ptrsize)
        self.avail_meta = next_int(ptrsize)
        self.avail_meta_count = next_int(size_tsize)
        self.avail_meta_area_count = next_int(size_tsize)
        self.meta_alloc_shift = next_int(size_tsize)
        self.meta_area_head = next_int(ptrsize)
        self.meta_area_tail = next_int(ptrsize)
        self.avail_meta_areas = next_int(ptrsize)

        assert len(size_classes) == 48

        for _ in range(len(size_classes)):
            cur_active = next_int(ptrsize)
            self.active.append(cur_active)

        for _ in range(len(size_classes)):
            cur_usage = next_int(size_tsize)
            self.usage_by_class.append(cur_usage)

        for _ in range(32):
            cur_seq = next_int(uint8size)
            self.unmap_seq.append(cur_seq)

        for _ in range(32):
            cur_bounce = next_int(uint8size)
            self.bounces.append(cur_bounce)

        self.seq = next_int(uint8size)

        # Adjust for alignment
        cur_offset += ptrsize - uint8size

        self.brk = next_int(ptrsize)

        assert cur_offset == self.sizeof

    def looks_valid(self) -> bool:
        """
        Returns true if this object looks like a valid `struct malloc_context` object
        describing an initialized heap. False otherwise.

        This is used by `class Mallocng` to find the correct ctx object.

        We consider it invalid if the heap reads as uninitialized because:
        1. Performing this check filters out invalid ctx objects very well.
        2. When musl is dynmically linked, due to the ld donation logic,
           the heap will usually be initialized before the start of main().
        """
        if self.init_done != 1:
            return False

        if self.secret <= 0x0000FFFFFFFFFFFF:
            # 1 / 65536 chance this is a false negative.
            return False

        return True


class Mallocng(pwndbg.aglib.heap.heap.MemoryAllocator):
    """
    Tracks the allocator state.
    By leveraging the __malloc_context symbol.

    Import this singleton class like:
    from pwndbg.aglib.heap.mallocng import ng

    and make sure that you have run ng.init_if_needed()
    before you used the object.
    """

    def __init__(self) -> None:
        self.finished_init: bool = False

        self.ctx_addr: int = 0
        self.ctx: MallocContext | None = None
        self.has_internal_syms: bool = False

    def init_if_needed(self) -> bool:
        """
        We want this class to be a singleton, but also we can't
        initialize it as soon as pwndbg is loaded.

        Users of the object are responsible for calling this to
        make sure the object is initialized. This also ensures
        our view of the heap is up-to-date.

        Returns:
            True if this object is successfully initialized (whether
            now or before). False otherswise. If this returns False
            you may not use this object for heap operations.
        """
        if self.finished_init:
            # Whoever called init_if_needed() needs to use the Mallocng
            # class, which needs an up-to-date view of __malloc_context,
            # so we will update it here.
            assert self.ctx is not None, "Init was finished but self.ctx is not initialized?"
            self.ctx.load()
            return True

        self.ctx_addr = 0
        self.ctx = None
        self.has_internal_syms = False

        # We will go in optimistically, and let set_ctx_addr() potentially
        # prove us wrong.
        self.finished_init = True
        self.set_ctx_addr()

        # If we failed, we will try again next time.

        return self.finished_init

    def set_ctx_addr(self) -> None:
        """
        Find where the __malloc_context global symbol is. Try using debug information,
        but if it isn't available try using a heuristic.
        """
        uint64size = pwndbg.aglib.typeinfo.uint64.sizeof

        ctx_addr_maybe: int | None = pwndbg.aglib.symbol.lookup_symbol_addr("__malloc_context")
        if ctx_addr_maybe is not None:
            self.ctx_addr = ctx_addr_maybe
            self.has_internal_syms = True
            self.ctx = MallocContext(self.ctx_addr)
            return

        # No debug information :(
        self.ctx_addr = 0
        self.has_internal_syms = False

        # We will find the __malloc_context object by searching memory for
        # the secret.
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/malloc.c#L50
        # Extract the secret first.
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/glue.h#L49
        at_random = int(pwndbg.auxv.get()["AT_RANDOM"])
        secret = memory.read(at_random + 8, uint64size)

        secret_matches = list(
            pwndbg.search.search(secret, executable=False, writable=True, aligned=uint64size)
        )

        # There are going to be multiple matches. We don't
        # want those on the stack (actual AT_RANDOM) or heap
        # (structures copying the secret). We want it either from the libc.so
        # mapping (if musl is dynamically linked) or the executable's
        # mapping (if musl is statically linked).
        possible: list[tuple[int, str]] = []
        thread_stacks = pwndbg.aglib.stack.get().values()

        for sm in secret_matches:
            if any(sm in stack_page for stack_page in thread_stacks):
                continue

            mapping = pwndbg.aglib.vmmap.find(sm)
            if mapping is None:
                continue
            if "[heap" in mapping.objfile:
                continue

            possible.append((sm, mapping.objfile))

        if not possible:
            print(message.error("Couldn't find __malloc_context, even with heuristic."))
            print(message.error("Musl mallocng commands will not work.\n"))
            self.ctx_addr = 0
            self.finished_init = False
            return

        known_invalid: set[int] = set()

        if pwndbg.dbg.selected_inferior().is_dynamically_linked():
            for addr, mapname in possible:
                # Unlike glibc, for musl the libc and ld are loaded as one object file.
                # On some distro's the object file shows up as libc
                # (e.g. /usr/lib/musl/lib/libc.so on archlinux) and on some as ld
                # (e.g. /usr/lib/ld-musl-x86_64.so.1 on fedora).
                if mapname.endswith("libc.so") or "/ld-musl-" in mapname:
                    maybe_ctx = MallocContext(addr)
                    if maybe_ctx.looks_valid():
                        self.ctx_addr = addr
                        self.ctx = maybe_ctx
                        return
                    known_invalid.add(addr)

            for addr, mapname in possible:
                if addr in known_invalid:
                    continue

                # It is probably better for this matching to be overly eager than overly restrictive,
                # we can fix it later if we ever actually encounter false positives.
                # The .startswith() logic is there for CTF-type setups when the libc/ld is in the same folder as the binary.
                if (
                    "/libc" in mapname
                    or "/ld-" in mapname
                    or mapname.startswith("libc")
                    or mapname.startswith("ld-")
                ):
                    maybe_ctx = MallocContext(addr)
                    if maybe_ctx.looks_valid():
                        self.ctx_addr = addr
                        self.ctx = maybe_ctx
                        return
                    known_invalid.add(addr)

            print(
                message.warn(
                    "Couldn't find __malloc_context in a 'libc' or 'ld-' mapping, trying elsewhere."
                )
            )
        else:
            # Statically linked.
            # TODO: We should find the Executable Object in the mappings
            # and use that to determine which match is correct. Not sure
            # how to do that though so fall through for now.
            pass

        for addr, mapname in possible:
            if addr not in known_invalid:
                maybe_ctx = MallocContext(addr)
                if maybe_ctx.looks_valid():
                    self.ctx_addr = addr
                    self.ctx = maybe_ctx
                    break

        if self.ctx_addr == 0 or self.ctx is None:
            print(
                message.error(
                    "Couldn't find a valid looking __malloc_context, even with heuristic."
                )
            )
            print(
                message.error(
                    "Musl mallocng commands will not work. Is the allocator initialized?\n"
                )
            )
            self.ctx_addr = 0
            self.ctx = None
            self.finished_init = False
            return

        # Tell the user we found __malloc_context but in an unexpected place.
        if pwndbg.dbg.selected_inferior().is_dynamically_linked():
            print(
                message.warn(
                    f"Found a match @ {self.ctx_addr:#x}. A bit suspicious but we will roll with it.\n"
                )
            )

    def find_slot(
        self, address: int, metadata: bool = False, shallow: bool = False
    ) -> tuple[GroupedSlot | None, Slot | None]:
        """
        Get the slot which contains this address.

        We say a slot "contains" an address, if the address is in
        [start, start + stride). Thus, this will match the previous
        slot if you provide the address of the header inband metadata
        of a slot.

        If `metadata` is True, then we check [start - IB, end) for
        containment.

        If `shallow` is True, return the biggest slot which contains this
        address. The group that owns this slot will not be a nested group.

        Returns (None, None) if nothing is found.
        """
        if self.ctx is None:
            raise AssertionError("You didn't initialize ng.")

        metadata_offset = IB if metadata else 0
        # The group which contains a slot which contains `address`.
        hit_group: Group | None = None

        meta_area_addr = self.ctx.meta_area_head
        while meta_area_addr:
            try:
                meta_area = MetaArea(meta_area_addr)
            except pwndbg.dbg_mod.Error as e:
                # Can't get `next` if the main_area is corrupted.
                print(
                    message.error(
                        f"Mallocng.containing: Could not read meta_area ({e}), returning early."
                    )
                )
                return None, None

            # Iterate over all metas in the meta_area.
            for i in range(meta_area.nslots):
                try:
                    meta = Meta(meta_area.at_index(i))
                    if not meta.mem:
                        # Skip unused metas.
                        continue
                except pwndbg.dbg_mod.Error as e:
                    print(
                        message.error(
                            f"Mallocng.containing: Could not read/parse meta.({e}), skipping it.."
                        )
                    )
                    continue

                try:
                    group = Group(meta.mem)
                    group.set_meta(meta)

                    valid_start = group.storage - metadata_offset
                    group_end = group.addr + group.group_size

                    # Check if our address is inside one of
                    # the group's slots.
                    if valid_start <= address < group_end:
                        # Yes it is!
                        hit_group = group
                        break
                except pwndbg.dbg_mod.Error as e:
                    print(
                        message.error(
                            "Mallocng.containing: Could not read/parse meta at"
                            f" {hex(meta.addr)} ({e}), skipping it.."
                        )
                    )
                    continue

            if hit_group:
                break

            meta_area_addr = meta_area.next

        if hit_group is None:
            return None, None

        # Need to read memory for the .contains_group() check.
        hit_slot: Slot | None = None
        # Contains extra information.
        hit_grouped_slot: GroupedSlot | None = None

        if shallow:
            backup_addr = hit_group.addr
            try:
                # Go up instead of recursing.
                hit_group = hit_group.meta.root_group()
                slot_idx = (
                    address - (hit_group.storage - metadata_offset)
                ) // hit_group.meta.stride
                hit_grouped_slot = GroupedSlot(hit_group, slot_idx)
                hit_slot = Slot.from_start(hit_grouped_slot.start)
                return hit_grouped_slot, hit_slot
            except pwndbg.dbg_mod.Error as e:
                print(
                    message.error(
                        "Mallocng.containing: Failed reading memory while traversing"
                        f" parent groups to satisfy shallow=True.\n{e}.\n"
                        f"The initial match was for group @ {backup_addr}.\n"
                    )
                )
                return None, None

        try:
            # Recursively go into deeper nested groups until we find a slot
            # which doesn't house a group.
            while hit_slot is None or hit_slot.contains_group():
                valid_start = hit_group.storage - metadata_offset

                if address < valid_start:
                    # Bleh, the address is in the group's header
                    # (or the first slot's start header). What to do?
                    if hit_slot is not None:
                        # If we are already in some slot, just return
                        # that slot since we can't look any deeper.
                        break
                    # We are in no slot i.e. we are in the header of a
                    # top level group (either mmap()ed or donated).
                    # We could return *some* information to the callee
                    # but alas, let's be technically correct.
                    return None, None

                # Calculate the correct inner slot.
                slot_idx = (address - valid_start) // hit_group.meta.stride

                hit_grouped_slot = GroupedSlot(hit_group, slot_idx)
                hit_slot = Slot.from_start(hit_grouped_slot.start)

                # If the slot is not allocated, we know that we for sure can't
                # recurse deeper.
                if hit_grouped_slot.slot_state != SlotState.ALLOCATED:
                    break

                # Maybe there is a group inside this slot!
                hit_group = Group(hit_slot.p)

            return hit_grouped_slot, hit_slot

        except pwndbg.dbg_mod.Error as e:
            print(
                message.error(
                    "Mallocng.containing: Failed reading memory while traversing"
                    f" nested groups: {e}.\nReturning last valid slot."
                )
            )
            # Could be None.
            return hit_grouped_slot, hit_slot

    @override
    def containing(self, address: int, metadata: bool = False, shallow: bool = False) -> int:
        """
        Same as find_slot() but returns only the `start` address of the slot, or zero
        if no slot is found.
        """
        found, _ = self.find_slot(address, metadata, shallow)
        if found is None:
            return 0
        return found.start

    def get_free_metas(self) -> dict[int, tuple[int, Meta]]:
        """
        Get all free metas by traversing the ctx.free_meta_head doubly
        linked list. Map them to their index in the list.

        Raises:
            pwndbg.dbg_mod.Error: If some meta cannot be read or is
                corrupted.

        Returns:
            A dictionary that maps: meta address -> (meta index in list, Meta object).
        """
        if self.ctx is None:
            raise AssertionError("You didn't initialize ng.")

        if self.ctx.free_meta_head == 0:
            return {}

        # A free meta gets completely cleared.
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L111
        # Except for the prev/next pointers which add it to the ctx.free_meta_head chain.
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/meta.h#L85
        # We could check for everything, but mallocng doesn't so lets not.

        start_meta = Meta(self.ctx.free_meta_head)
        meta_dict: dict[int, tuple[int, Meta]] = {start_meta.addr: (0, start_meta)}

        idx = 1
        cur_meta = Meta(start_meta.next)
        while cur_meta.addr != start_meta.addr:
            meta_dict[cur_meta.addr] = (idx, cur_meta)

            idx += 1
            cur_meta = Meta(cur_meta.next)

        return meta_dict

    def meta_is_avail(self, addr: int) -> bool:
        """
        Checks whether a meta is available.
        """
        if self.ctx is None:
            raise AssertionError("You didn't initialize ng.")

        # It seems all available metas are contiguous.
        # https://elixir.bootlin.com/musl/v1.2.5/source/src/malloc/mallocng/malloc.c#L109
        return (
            self.ctx.avail_meta
            <= addr
            < (self.ctx.avail_meta + Meta.sizeof() * self.ctx.avail_meta_count)
        )


ng = Mallocng()



from __future__ import annotations

import argparse

from elftools.elf.elffile import ELFFile

import pwndbg.aglib
import pwndbg.aglib.file
import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.color.memory as mem_color
import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""Prints the section mappings contained in the ELF header.
If binary not start or use --no-rebase, the section permission based on section flags.
""",
)

parser.add_argument(
    "-R",
    "--no-rebase",
    help="Print the non-rebased section address. ",
    action="store_true",
    default=False,
    dest="no_rebase",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.LINUX,
    examples="""
elfsections
elfsections --no-rebase
""",
)
@pwndbg.commands.OnlyWithFile
def elfsections(no_rebase: bool) -> None:
    local_path = pwndbg.aglib.file.get_proc_exe_file()

    bin_base_addr = 0
    # Get the binary base address, for rebase the section address if we need.
    if pwndbg.aglib.proc.alive():
        bin_base_addr = pwndbg.aglib.proc.binary_base_addr()

    __SH_WRITE = 1 << 0
    __SH_ALLOC = 1 << 1
    __SH_EXEC = 1 << 2

    with open(local_path, "rb") as f:
        elffile = ELFFile(f)
        sections = []
        for section in elffile.iter_sections():
            start = section["sh_addr"]
            privilege = section["sh_flags"]

            # Don't print sections that aren't mapped into memory
            # test in go sample: "gosample.x64/86", some sections have the address but don't have the SH_ALLOC flags
            if not (privilege & __SH_ALLOC):
                continue

            size = section["sh_size"]

            # rebase the address if we need
            if not no_rebase:
                start = bin_base_addr + start if start < bin_base_addr else start

            sections.append([start, start + size, size, section.name, privilege])

        sections.sort()

        # print legend
        print(mem_color.legend())

        # table header
        print(f"{'Start':>18} {'End':>18} {'Perm':>8} {'Size':>10}  {'Name':<}")

        # if the binary is started, use the memory permission for the coloring
        if pwndbg.aglib.proc.alive() and not no_rebase:
            for start, end, size, name, privilege in sections:
                page = pwndbg.aglib.vmmap.find(start)

                privilege_str = "R" if page.read else "-"
                privilege_str += "W" if page.write else "-"
                privilege_str += "X" if page.execute else "-"

                print(
                    mem_color.get(
                        start,
                        text=f"{start:>#18x} {end:>#18x} {privilege_str:>8} {size:>#10x}  {name:<}",
                    )
                )
        else:
            # if the binary is not start, use the section flags for the coloring.
            for start, end, size, name, privilege in sections:
                color = mem_color.c.rodata
                privilege_str = "R"

                if privilege & __SH_WRITE:
                    privilege_str += "W"
                    color = mem_color.c.data
                else:
                    privilege_str += "-"

                if privilege & __SH_EXEC:
                    privilege_str += "X"
                    color = mem_color.c.code
                else:
                    privilege_str += "-"

                print(color(f"{start:>#18x} {end:>#18x} {privilege_str:>8} {size:>#10x}  {name:<}"))


@pwndbg.commands.Command(
    "Prints any symbols found in the .got.plt section if it exists.",
    category=CommandCategory.LINUX,
)
@pwndbg.commands.OnlyWithFile
def gotplt() -> None:
    print_symbols_in_section(".got.plt", "@got.plt")


# These are all the section names associated with PLTs.
# .plt.sec and .plt.bnd are associated with control flow transfer integrity.
# These are derived from this list that GDB recognizes: https://github.com/bminor/binutils-gdb/blob/38d726a24c1a85abdb606e7ab6cefad17872aad7/bfd/elf64-x86-64.c#L5775-L5780
PLT_SECTION_NAMES = (".plt", ".plt.sec", ".plt.got", ".plt.bnd")

parser = argparse.ArgumentParser(
    description="Prints any symbols found in Procedure Linkage Table sections if any exist.",
)

parser.add_argument(
    "-a",
    "--all-symbols",
    help="Print all symbols, not just those that end in @plt",
    action="store_true",
    default=False,
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.LINUX,
)
@pwndbg.commands.OnlyWithFile
def plt(all_symbols: bool = False) -> None:
    local_path = pwndbg.aglib.file.get_proc_exe_file()

    bin_base_addr = 0
    # If we started the binary and it has PIE, rebase it
    if pwndbg.aglib.proc.alive():
        bin_base_addr = pwndbg.aglib.proc.binary_base_addr()

    # List of (Section name, start_addr, end_addr)
    sections_found: list[tuple[str, int, int]] = []

    with open(local_path, "rb") as f:
        elffile = ELFFile(f)

        for section_name in PLT_SECTION_NAMES:
            section = elffile.get_section_by_name(section_name)

            if section:
                start: int = section["sh_addr"]
                size: int = section["sh_size"]

                if start is None:
                    continue

                end = start + size

                # Rebase the start and end addresses if needed
                if start < bin_base_addr:
                    start += bin_base_addr
                    end += bin_base_addr

                sections_found.append((section_name, start, end))

    # Sort by the start address so we print from lowest to highest
    sections_found.sort(key=lambda x: x[1])

    for section_name, start, end in sections_found:
        symbols = get_symbols_in_region(start, end, "" if all_symbols else "@plt")

        print(message.notice(f"Section {section_name} {start:#x} - {end:#x}:"))

        if not symbols:
            print(message.error(f"No symbols found in section {section_name}"))

        stuff: list[tuple[int, str]] = []

        for symbol, addr in symbols:
            stuff.append((addr, symbol))
            print(hex(int(addr)) + ": " + symbol)

    if len(sections_found) == 0:
        print(message.error("No .plt.* sections found"))


def get_section_bounds(section_name: str):
    local_path = pwndbg.aglib.file.get_proc_exe_file()

    with open(local_path, "rb") as f:
        elffile = ELFFile(f)

        section = elffile.get_section_by_name(section_name)

        if not section:
            return (None, None)

        start = section["sh_addr"]
        size = section["sh_size"]
        return (start, start + size)


def print_symbols_in_section(section_name, filter_text="") -> None:
    start, end = get_section_bounds(section_name)

    if start is None:
        print(message.error(f"Could not find section {section_name}"))
        return

    # If we started the binary and it has PIE, rebase it
    if pwndbg.aglib.proc.alive():
        bin_base_addr = pwndbg.aglib.proc.binary_base_addr()

        # Rebase the start and end addresses if needed
        if start < bin_base_addr:
            start += bin_base_addr
            end += bin_base_addr

    print(message.notice(f"Section {section_name} {start:#x}-{end:#x}:"))

    symbols = get_symbols_in_region(start, end, filter_text)

    if not symbols:
        print(message.error(f"No symbols found in section {section_name}"))

    for symbol, addr in symbols:
        print(hex(int(addr)) + ": " + symbol)


def get_symbols_in_region(start: int, end: int, filter_text="") -> list[tuple[str, int]]:
    symbols: list[tuple[str, int]] = []
    ptr_size = pwndbg.aglib.typeinfo.pvoid.sizeof
    addr = start
    while addr < end:
        name = pwndbg.aglib.symbol.resolve_addr(addr)
        if name and "+" not in name and filter_text in name:
            symbols.append((name, addr))
        addr += ptr_size

    return symbols



from __future__ import annotations

import functools
from collections.abc import Iterator

import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo
import pwndbg.dbg_mod
from pwndbg.aglib.kernel.macros import container_of
from pwndbg.aglib.kernel.macros import for_each_entry

NFPROTO_INET = 1
NFPROTO_IPV4 = 2
NFPROTO_ARP = 3
NFPROTO_NETDEV = 5
NFPROTO_BRIDGE = 7
NFPROTO_IPV6 = 10

nftables_table_family = {
    "inet": NFPROTO_INET,
    "ip": NFPROTO_IPV4,
    "arp": NFPROTO_ARP,
    "netdev": NFPROTO_NETDEV,
    "bridge": NFPROTO_BRIDGE,
    "ip6": NFPROTO_IPV6,
}


def catch_error(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except (pwndbg.dbg_mod.Error, UnicodeDecodeError) as e:
            return f"<unavailable: {e}>"

    return wrapper


class NftFields:
    @catch_error
    def __getattr__(self, name: str):
        t = self.__annotations__.get(name)
        if t == "str":
            return self._addr[name].string()
        if t == "int":
            return int(self._addr[name])

        raise AttributeError(f"'{t}' this type hint is not supported on field '{name}'")

    @classmethod
    def get_hook_list_dev_names(cls, hook_list: pwndbg.dbg_mod.Value) -> list[str]:
        devs = []
        for hook in for_each_entry(hook_list, "struct nft_hook", "list"):
            dev_name = hook["ops"]["dev"]["name"].string()
            devs.append(dev_name)
        return devs

    def print_fields(self, nested: int = 0, keys: list[str] = None):
        pad = " " * ((nested + 1) * 2)
        if keys is None:
            keys = self.__annotations__.keys()
        for key in keys:
            val = getattr(self, key)
            if hasattr(val, "nested_print"):
                print(f"{pad}{key}:")
                val.nested_print(nested=nested + 1)
                continue

            print(f"{pad}{key}: {val}")


class Expr:
    expr_name: str

    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct nft_expr *

    @property
    def expr_name(self) -> str:
        return self._addr["ops"]["type"]["name"].string()

    def print_expr_iptables(self, expr_name: str):
        typel = pwndbg.aglib.typeinfo.lookup_types(f"struct xt_{expr_name}")
        expr = pwndbg.aglib.memory.get_typed_pointer(typel, int(self._addr["ops"]["data"]))

        if expr_name == "match":
            size = int(expr["matchsize"])
        elif expr_name == "target":
            size = int(expr["targetsize"])
        else:
            assert False

        try:
            info = pwndbg.aglib.memory.read(int(self._addr["data"].address), size)
        except pwndbg.dbg_mod.Error as e:
            info = bytearray(f"<unavailable: {e}>".encode())
        print(
            {
                "name": expr["name"].string(),
                "rev": int(expr["revision"]),
                "size": size,
                "info": info,
            }
        )

    def print_expr_nftables(self, expr_name: str):
        typel = pwndbg.aglib.typeinfo.lookup_types(
            f"struct nft_{expr_name}",
            f"struct nft_{expr_name}_expr",
        )
        expr = pwndbg.aglib.memory.get_typed_pointer(typel, int(self._addr["data"].address))
        print(expr.dereference().value_to_human_readable())

    def print(self, print_nested: bool = True):
        expr_name = self.expr_name
        addr = int(self._addr["data"].address)
        print(f"{expr_name} @ 0x{addr:x}")

        if not print_nested:
            return

        if int(self._addr["ops"]["dump"].address) == 0:
            # No dump method, skip
            return

        if expr_name in ("match", "target"):
            self.print_expr_iptables(expr_name)
        else:
            self.print_expr_nftables(expr_name)


class Rule(NftFields):
    handle: int  # NFTA_RULE_HANDLE
    userdata: bytearray  # NFTA_RULE_USERDATA

    def __init__(self, addr: pwndbg.dbg_mod.Value, chain: Chain):
        self._addr = addr  # struct nft_rule *
        self._chain = chain

    @property
    @catch_error
    def userdata(self) -> bytearray:
        if int(self._addr["udata"]) == 0:
            return bytearray(b"")

        ptr = int(self._addr["data"].address) + int(self._addr["dlen"])
        userdata = pwndbg.aglib.memory.get_typed_pointer("struct nft_userdata", ptr)
        userdata_str = pwndbg.aglib.memory.string(
            int(userdata["data"].address), int(userdata["len"])
        )
        return userdata_str

    @classmethod
    def find(
        cls,
        table_name: str,
        table_family: int,
        chain_name: str,
        rule_id: int,
        nsid: int | None = None,
    ) -> Iterator[Rule]:
        for nft in Chain.find(
            table_name=table_name, table_family=table_family, chain_name=chain_name, nsid=nsid
        ):
            for rule in nft.iter_rules():
                if rule.handle == rule_id or rule_id is None:
                    yield rule

    def iter_exprs(self) -> Iterator[Expr]:
        start = int(self._addr["data"].address)
        end = start + int(self._addr["dlen"])
        current = start

        while current < end:
            ptr = pwndbg.aglib.memory.get_typed_pointer("struct nft_expr", current)
            yield Expr(ptr)
            current += int(ptr["ops"]["size"])

    def print(self, print_nested: bool = True):
        print()
        print(f"nft_rule @ 0x{int(self._addr):x}")
        self.print_fields()

        if not print_nested:
            return
        for expr in self.iter_exprs():
            expr.print()


class ChainHook(NftFields):
    hooknum: int  # NFTA_HOOK_HOOKNUM
    priority: int  # NFTA_HOOK_PRIORITY

    # custom types
    dev: str  # NFTA_HOOK_DEV
    devs: list[str]  # NFTA_HOOK_DEVS

    def __init__(self, parent: Chain):
        basechain = parent.basechain
        if basechain is None:
            self._addr = None
            self._parent = None
            return

        self._addr = basechain["ops"]
        self._parent = parent

    @property
    @catch_error
    def dev(self) -> str:
        devs = self.get_netdevs()
        if len(devs) == 1:
            return devs[0]
        return ""

    @property
    @catch_error
    def devs(self) -> list[str]:
        return self.get_netdevs()

    def is_netdev(self) -> bool:
        NF_INET_INGRESS = 5
        family = self._parent.table.family
        hooknum = self.hooknum
        return family == NFPROTO_NETDEV or (family == NFPROTO_INET and hooknum == NF_INET_INGRESS)

    def get_netdevs(self) -> list[str]:
        basechain = self._parent.basechain
        if basechain is None:
            return []

        if not self.is_netdev():
            return []

        return self.get_hook_list_dev_names(basechain["hook_list"])

    def nested_print(self, nested: int = 0):
        self.print_fields(nested=nested)


class Chain(NftFields):
    bound: int  # internal field
    genmask: int  # internal field

    handle: int  # NFTA_CHAIN_HANDLE
    use: int  # NFTA_CHAIN_USE
    flags: int  # NFTA_CHAIN_FLAGS
    name: str  # NFTA_CHAIN_NAME

    # custom types
    hook: ChainHook  # NFTA_CHAIN_HOOK
    table: Table  # NFTA_CHAIN_TABLE
    userdata: bytearray  # NFTA_CHAIN_USERDATA
    policy: int  # NFTA_CHAIN_POLICY
    type: str  # NFTA_CHAIN_TYPE
    # TODO: implement: counters # NFTA_CHAIN_COUNTERS chain->basechain->stats (struct nft_stats __percpu *stats;)

    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct nft_chain *
        self.hook = ChainHook(self)

    @property
    @catch_error
    def userdata(self) -> bytearray:
        udata = self._addr["udata"]
        if int(udata) == 0:
            return bytearray(b"")

        return pwndbg.aglib.memory.string(int(udata.address), int(self._addr["udlen"]))

    @property
    @catch_error
    def policy(self) -> int:
        return int(self.basechain["policy"])

    @property
    @catch_error
    def type(self) -> str:
        return self.basechain["type"]["name"].string()

    @property
    @catch_error
    def table(self) -> Table:
        return Table(self._addr["table"])

    @property
    @catch_error
    def basechain(self) -> pwndbg.dbg_mod.Value | None:
        NFT_CHAIN_BINDING = 1 << 2
        if self.flags & NFT_CHAIN_BINDING:
            return None

        return container_of(int(self._addr), "struct nft_base_chain", "chain")

    @classmethod
    def find(
        cls,
        table_family: int | None = None,
        table_name: str | None = None,
        chain_name: str | None = None,
        nsid: int | None = None,
    ) -> Iterator[Chain]:
        for nft in Table.find(table_name=table_name, table_family=table_family, nsid=nsid):
            for chain in nft.iter_chains():
                if chain_name is None or chain.name == chain_name:
                    yield chain

    def iter_rules(self) -> Iterator[Rule]:
        for rule in for_each_entry(self._addr["rules"], "struct nft_rule", "list"):
            yield Rule(rule, self)

    def print(self, print_nested: bool = True):
        print()
        print(f"nft_chain @ 0x{int(self._addr):x}")
        self.print_fields()

        if not print_nested:
            return
        for rule in self.iter_rules():
            rule.print()


class Set(NftFields):
    use: int  # internal field
    dead: int  # internal field
    genmask: int  # internal field
    field_count: int  # internal field
    num_exprs: int  # internal field

    name: str  # NFTA_SET_NAME
    flags: int  # NFTA_SET_FLAGS
    ktype: int  # NFTA_SET_KEY_TYPE
    klen: int  # NFTA_SET_KEY_LEN
    dtype: int  # NFTA_SET_DATA_TYPE
    dlen: int  # NFTA_SET_DATA_LEN
    policy: int  # NFTA_SET_POLICY
    timeout: int  # NFTA_SET_TIMEOUT
    gc_int: int  # NFTA_SET_GC_INTERVAL
    objtype: int  # NFTA_SET_OBJ_TYPE
    handle: int  # NFTA_SET_HANDLE

    # custom fields
    nelems: int  # internal field
    table: Table  # NFTA_SET_TABLE
    userdata: bytearray  # NFTA_SET_USERDATA
    desc_size: int  # NFTA_SET_DESC_SIZE
    desc_concat: int  # NFTA_SET_DESC_CONCAT
    # iter_exprs  # NFTA_SET_EXPR
    # iter_exprs  # NFTA_SET_EXPRESSIONS

    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct nft_set *

    @property
    @catch_error
    def nelems(self) -> int:
        return int(self._addr["nelems"]["counter"])

    @property
    @catch_error
    def desc_size(self) -> int:
        return int(self._addr["size"])

    @property
    @catch_error
    def desc_concat(self) -> list[int]:
        field_count = self.field_count
        if not (field_count > 1):
            return []

        out = []
        for i in range(field_count):
            val = int(self._addr["field_len"][i])  # todo: htonl?
            out.append(val)
        return out

    @property
    @catch_error
    def table(self) -> Table:
        return Table(self._addr["table"])

    @property
    @catch_error
    def userdata(self) -> bytearray:
        if int(self._addr["udata"]) == 0:
            return bytearray(b"")

        return pwndbg.aglib.memory.string(
            int(self._addr["udata"].address), int(self._addr["udlen"])
        )

    def iter_expr(self) -> Iterator[Expr]:
        for i in range(self.num_exprs):
            yield Expr(self._addr["exprs"][i])

    def iter_elems(self) -> Iterator[None]:
        # TODO: implement nft_get_set_elem
        raise NotImplementedError("todo implement nft_get_set_elem")

    def print(self, print_nested: bool = True):
        print()
        print(f"nft_set @ 0x{int(self._addr):x}")
        self.print_fields()

        for expr in self.iter_expr():
            expr.print()


# Object can only be used for:
# - nft_connlimit
# - nft_ct
# - nft_limit
# - nft_quota
# - nft_synproxy
# - nft_tunnel
class Object(NftFields):
    genmask: int  # internal field

    use: int  # NFTA_OBJ_USE
    handle: int  # NFTA_OBJ_HANDLE

    # custom fields
    table: Table  # NFTA_OBJ_TABLE
    name: str  # NFTA_OBJ_NAME
    type: int  # NFTA_OBJ_TYPE
    userdata: bytearray  # NFTA_OBJ_USERDATA
    # iter_data  # NFTA_OBJ_DATA

    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct nft_object *

    @property
    @catch_error
    def type(self) -> int:
        return int(self._addr["ops"]["type"]["type"])

    @property
    @catch_error
    def name(self) -> str:
        return self._addr["key"]["name"].string()

    @property
    @catch_error
    def table(self) -> Table:
        return Table(self._addr["key"]["table"])

    @property
    @catch_error
    def userdata(self) -> bytearray:
        if int(self._addr["udata"]) == 0:
            return bytearray(b"")

        return pwndbg.aglib.memory.string(
            int(self._addr["udata"].address), int(self._addr["udlen"])
        )

    def iter_data(self) -> Iterator[None]:
        # TODO: implement nft_object_dump
        raise NotImplementedError("todo implement nft_object_dump")

    def print(self, print_nested: bool = True):
        print()
        print(f"nft_object @ 0x{int(self._addr):x}")
        self.print_fields()


class FlowtableHook(NftFields):
    hooknum: int  # NFTA_FLOWTABLE_HOOK_NUM
    priority: int  # NFTA_FLOWTABLE_HOOK_PRIORITY
    devs: list[str]  # NFTA_FLOWTABLE_HOOK_DEVS

    def __init__(self, parent: Flowtable):
        self._addr = parent._addr

    @property
    @catch_error
    def hooknum(self) -> int:
        return int(self._addr["hooknum"])

    @property
    @catch_error
    def priority(self) -> int:
        return int(self._addr["data"]["priority"])

    @property
    @catch_error
    def devs(self) -> list[str]:
        return self.get_hook_list_dev_names(self._addr["hook_list"])

    def nested_print(self, nested: int = 0):
        self.print_fields(nested=nested)


class Flowtable(NftFields):
    genmask: int  # internal field

    name: str  # NFTA_FLOWTABLE_NAME
    use: int  # NFTA_FLOWTABLE_USE
    handle: int  # NFTA_FLOWTABLE_HANDLE

    # custom fields
    table: Table  # NFTA_FLOWTABLE_TABLE
    flags: int  # NFTA_FLOWTABLE_FLAGS
    hook: FlowtableHook  # NFTA_FLOWTABLE_HOOK

    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct nft_flowtable *
        self.hook = FlowtableHook(self)

    @property
    @catch_error
    def table(self) -> Table:
        return Table(self._addr["table"])

    @property
    @catch_error
    def flags(self) -> int:
        return int(self._addr["data"]["flags"])

    def print(self, print_nested: bool = True):
        print()
        print(f"nft_flowtable @ 0x{int(self._addr):x}")
        self.print_fields()


class Table(NftFields):
    family: int  # internal field
    genmask: (
        int  # internal nft transaction number (maybe useful to checking errors in commit phase)
    )
    name: str  # NFTA_TABLE_NAME
    handle: int  # NFTA_TABLE_HANDLE
    use: int  # NFTA_TABLE_USE
    flags: int  # NFTA_TABLE_FLAGS
    nlpid: int  # NFTA_TABLE_OWNER

    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct nft_table *

    @classmethod
    def find(
        cls,
        table_name: str | None = None,
        table_family: int | None = None,
        nsid: int | None = None,
    ) -> Iterator[Table]:
        nft = Nftables.find(nsid=nsid)
        if nft is None:
            return
        for table in nft.iter_tables():
            if (table.family == table_family or table_family is None) and (
                table.name == table_name or table_family is None
            ):
                yield table

    def iter_chains(self) -> Iterator[Chain]:
        for chain in for_each_entry(self._addr["chains"], "struct nft_chain", "list"):
            yield Chain(chain)

    def iter_sets(self) -> Iterator[Set]:
        for nft_set in for_each_entry(self._addr["sets"], "struct nft_set", "list"):
            yield Set(nft_set)

    def iter_flowtables(self) -> Iterator[Flowtable]:
        for flowtable in for_each_entry(self._addr["flowtables"], "struct nft_flowtable", "list"):
            yield Flowtable(flowtable)

    def iter_objects(self) -> Iterator[Object]:
        for nft_object in for_each_entry(self._addr["objects"], "struct nft_flowtable", "list"):
            yield Object(nft_object)

    def nested_print(self, nested: int = 0):
        self.print_fields(nested=nested, keys=["name"])

    def print(self, print_nested: bool = True):
        print()
        print(f"nft_table @ 0x{int(self._addr):x}")
        self.print_fields()

        if not print_nested:
            return
        for chain in self.iter_chains():
            chain.print()


class Nftables:
    def __init__(self, addr: pwndbg.dbg_mod.Value):
        self._addr = addr  # struct net *

    @classmethod
    def find(cls, nsid: int | None = None) -> Nftables | None:
        if nsid is None:
            addr = get_init_net_namespace()
        else:
            # See kernel: `struct net *get_net_ns_by_id(const struct net *net, int id)`
            raise NotImplementedError("todo implement search netns by id")

        return Nftables(addr)

    def iter_tables(self) -> Iterator[Table]:
        for table in for_each_entry(self._addr["nft"]["tables"], "struct nft_table", "list"):
            yield Table(table)

    def print(self):
        for table in self.iter_tables():
            table.print()


def get_init_net_namespace() -> pwndbg.dbg_mod.Value:
    return pwndbg.aglib.symbol.lookup_symbol("init_net")


# TODO: List what is missing:
# - pretty printing expressions
# - listing nested expressions eg. in "dynset" expression or expresions inside setelem
# - bug with printing "counter", because there are percpu struct
# - printing/parser xtable (iptables) expressions
# - missing listing set elements: `knft-list-set-elements [-nsid <nsid>] <table_family> <table_name> <set_id>`
# - missing info about ebpf nft hooks
# - listing network namespaces
# - printing netns tables



from __future__ import annotations

import argparse
import shutil

import pwndbg.aglib.onegadget
import pwndbg.aglib.proc
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.libc
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""
Find gadgets which single-handedly give code execution.

Uses the onegadget tool by david942j.
""",
)

parser.add_argument("--show-unsat", help="Show unsatisfiable gadgets.", action="store_true")
parser.add_argument("--no-unknown", help="Do not show unknown gadgets.", action="store_true")
parser.add_argument("-v", "--verbose", help="Show verbose output.", action="store_true")


@pwndbg.commands.Command(parser, category=CommandCategory.LINUX)
@pwndbg.aglib.proc.OnlyWithArch(["x86-64", "i386", "aarch64"])
@pwndbg.commands.OnlyWhenRunning
def onegadget(show_unsat: bool = False, no_unknown: bool = False, verbose: bool = False) -> None:
    if not shutil.which("one_gadget"):
        print(
            message.error("Could not find one_gadget. Please ensure it's installed and in $PATH.")
        )
        return

    which_libc = pwndbg.libc.which()
    if which_libc != pwndbg.libc.LibcType.GLIBC:
        if which_libc == pwndbg.libc.LibcType.UNKNOWN:
            print(message.error("Could not find libc. Please ensure it's loaded."))
        else:
            print(
                message.error(
                    f"Currently active libc is {which_libc.value}, but onegadget only supports glibc."
                )
            )
        return

    path = pwndbg.libc.filepath()
    print(f"Using libc: {message.hint(path)}")
    print()

    gadgets_count = pwndbg.aglib.onegadget.find_gadgets(show_unsat, no_unknown, verbose)
    for result, count in gadgets_count.items():
        print(f"Found {message.hint(count)} {result} gadgets.")
    if not gadgets_count[pwndbg.aglib.onegadget.SAT] and not show_unsat:
        print(
            message.warn(
                "No valid gadgets found, you might want to run with --show-unsat again to check unsatisfiable gadgets.\n"
                "To see why they are unsatisfiable, you might want to run with -v or --verbose."
            )
        )



from __future__ import annotations

from collections.abc import Callable

from capstone6pwndbg.sparc import *  # noqa: F403
from typing_extensions import override

import pwndbg.aglib.disasm.arch
from pwndbg.aglib.disasm.instruction import ALL_JUMP_GROUPS
from pwndbg.aglib.disasm.instruction import InstructionCondition
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.emu.emulator import Emulator

# Instruction groups for future use
SPARC_LOAD_INSTRUCTIONS = {
    SPARC_INS_LDUB: 1,
    SPARC_INS_LDSB: 1,
    SPARC_INS_LDUH: 2,
    SPARC_INS_LDSH: 2,
    SPARC_INS_LD: 4,
    SPARC_INS_LDD: 8,
}

SPARC_STORE_INSTRUCTIONS = {
    SPARC_INS_STB: 1,
    SPARC_INS_STH: 2,
    SPARC_INS_ST: 4,
    SPARC_INS_STD: 8,
}

SPARC_CONDITIONAL_BRANCHES = {
    SPARC_INS_B,  # This is only conditional if .cc != SPARC_CC_ICC_A
    SPARC_INS_ALIAS_BE,
    SPARC_INS_ALIAS_BNE,
    SPARC_INS_ALIAS_BNEG,
    SPARC_INS_ALIAS_BPOS,
    SPARC_INS_ALIAS_BVS,
    SPARC_INS_ALIAS_BVC,
    SPARC_INS_ALIAS_BCS,
    SPARC_INS_ALIAS_BCC,
    SPARC_INS_ALIAS_BL,
    SPARC_INS_ALIAS_BLE,
    SPARC_INS_ALIAS_BG,
    SPARC_INS_ALIAS_BGE,
    SPARC_INS_ALIAS_BLEU,
    SPARC_INS_ALIAS_BGU,
}

# CCR BITS:
# 0 = C (Carry)
# 1 = V (Overflow)
# 2 = Z (Zero)
# 3 = N (Negative)
CCR_C_MASK = 1 << 0
CCR_V_MASK = 1 << 1
CCR_Z_MASK = 1 << 2
CCR_N_MASK = 1 << 3

# Key is ICC code
# lambda parameters: (current ccr value)
# Source of conditions: https://arcb.csc.ncsu.edu/~mueller/codeopt/codeopt00/notes/condbranch.html
ICC_CONDITION_RESOLVERS: dict[int, Callable[[int], bool]] = {
    SPARC_CC_ICC_NE: lambda ccr: not (ccr & CCR_Z_MASK),
    SPARC_CC_ICC_E: lambda ccr: bool(ccr & CCR_Z_MASK),
    SPARC_CC_ICC_G: lambda ccr: not ((ccr & CCR_N_MASK) ^ (ccr & CCR_V_MASK))
    and not (ccr & CCR_Z_MASK),
    SPARC_CC_ICC_LE: lambda ccr: bool((ccr & CCR_N_MASK) ^ (ccr & CCR_V_MASK))
    or bool(ccr & CCR_Z_MASK),
    SPARC_CC_ICC_GE: lambda ccr: not ((ccr & CCR_N_MASK) ^ (ccr & CCR_V_MASK)),
    SPARC_CC_ICC_L: lambda ccr: bool((ccr & CCR_N_MASK) ^ (ccr & CCR_V_MASK)),
    SPARC_CC_ICC_GU: lambda ccr: not (ccr & CCR_Z_MASK) and not (ccr & CCR_C_MASK),
    SPARC_CC_ICC_LEU: lambda ccr: bool(ccr & CCR_Z_MASK) or bool(ccr & CCR_C_MASK),
    SPARC_CC_ICC_CC: lambda ccr: not (ccr & CCR_C_MASK),
    SPARC_CC_ICC_CS: lambda ccr: bool(ccr & CCR_C_MASK),
    SPARC_CC_ICC_POS: lambda ccr: not (ccr & CCR_N_MASK),
    SPARC_CC_ICC_NEG: lambda ccr: bool(ccr & CCR_N_MASK),
    SPARC_CC_ICC_VC: lambda ccr: not (ccr & CCR_V_MASK),
    SPARC_CC_ICC_VS: lambda ccr: bool(ccr & CCR_V_MASK),
}


class SparcDisassemblyAssistant(pwndbg.aglib.disasm.arch.DisassemblyAssistant):
    @override
    def _condition(self, instruction: PwndbgInstruction, emu: Emulator) -> InstructionCondition:
        if instruction.id in SPARC_CONDITIONAL_BRANCHES:
            cc_field = instruction.cs_insn.cc_field

            if cc_field == SPARC_CC_FIELD_ICC:
                cc = instruction.cs_insn.cc

                # This indicates it is an unconditional branch
                if cc == SPARC_CC_ICC_A:
                    instruction.declare_is_unconditional_jump = True
                    return InstructionCondition.UNDETERMINED

                ccr = self._read_register_name(instruction, "ccr", emu)

                if ccr is None:
                    return InstructionCondition.UNDETERMINED

                conditional = ICC_CONDITION_RESOLVERS.get(cc, lambda *a: None)(ccr)

                if conditional is None:
                    return InstructionCondition.UNDETERMINED

                return InstructionCondition.TRUE if conditional else InstructionCondition.FALSE

        return InstructionCondition.UNDETERMINED

    @override
    def _resolve_target(self, instruction: PwndbgInstruction, emu: Emulator | None):
        if bool(instruction.groups & ALL_JUMP_GROUPS):
            instruction.causes_branch_delay = True

        return super()._resolve_target(instruction, emu)



from __future__ import annotations

import os
import random
import string
import subprocess
import sys
import tempfile

from pt.machine import Machine
from pt.pt import PageTableDump
from pt.pt_aarch64_parse import PT_Aarch64_Backend
from pt.pt_riscv64_parse import PT_RiscV64_Backend
from pt.pt_x86_64_parse import PT_x86_64_Backend

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.kernel
import pwndbg.aglib.kernel.paging
import pwndbg.aglib.memory
import pwndbg.aglib.qemu
import pwndbg.color.message as message
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.memory
from pwndbg.lib.memory import Page


class KernelVmmap:
    def __init__(self, pages: tuple[Page, ...]):
        self.pages = pages
        self.sections: tuple[tuple[str, int], ...] = None
        self.pi = pwndbg.aglib.kernel.arch_paginginfo()
        if self.pi:
            self.sections = self.pi.markers()
        self.adjust()

    def get_name(self, addr: int) -> str | None:
        if addr is None or self.sections is None:
            return None
        for i in range(len(self.sections) - 1):
            name, cur = self.sections[i]
            _, next = self.sections[i + 1]
            if cur is None or next is None or name is None:
                continue
            if cur <= addr < next:
                return name
        return None

    def adjust(self) -> None:
        if self.pi is None or self.pages is None or len(self.pages) == 0:
            return
        for i, page in enumerate(self.pages):
            name = self.get_name(page.start)
            if name is not None:
                page.objfile = name
        self.handle_user_pages()
        self.pi.handle_kernel_pages(self.pages)
        self.handle_offsets()

    def handle_user_pages(self) -> None:
        base_offset = self.pages[0].start
        for i in range(len(self.pages)):
            page = self.pages[i]
            if page.objfile != self.pi.USERLAND:
                break
            diff = page.start - base_offset
            if diff > 0x100000:
                if diff > 0x100000000000:
                    if page.execute:
                        page.objfile = "userland [library]"
                    elif page.rw:
                        page.objfile = "userland [stack]"
                else:
                    page.objfile = "userland [heap]"
            else:
                # page.objfile += f"_{hex(i)[2:]}"
                base_offset = page.start

    def handle_offsets(self) -> None:
        prev_objfile, base = "", 0
        for page in self.pages:
            # the check on KERNELRO is to make getting offsets for symbols such as `init_creds` more convinient
            if page.objfile != self.pi.KERNELRO and prev_objfile != page.objfile:
                prev_objfile = page.objfile
                base = page.start
            page.offset = page.start - base
            if len(hex(page.offset)) > 9:
                page.offset = 0


# Most of QemuMachine code was inherited from gdb-pt-dump thanks to Martin Radev (@martinradev)
# on the MIT license, see:
# https://github.com/martinradev/gdb-pt-dump/blob/21158ac3f9b36d0e5e0c86193e0ef018fc628e74/pt_gdb/pt_gdb.py#L11-L80
class QemuMachine(Machine):
    def __init__(self):
        super().__init__()
        self.file = None
        self.pid = QemuMachine.get_qemu_pid()
        self.file = os.open(f"/proc/{self.pid}/mem", os.O_RDONLY)

    def __del__(self):
        if self.file:
            os.close(self.file)

    @staticmethod
    def search_pids_for_file(pids: list[str], filename: str) -> str | None:
        for pid in pids:
            fd_dir = f"/proc/{pid}/fd"
            try:
                for fd in os.listdir(fd_dir):
                    if os.readlink(f"{fd_dir}/{fd}") == filename:
                        return pid
            except FileNotFoundError:
                # Either the process has gone or fds are changing, not our pid
                pass
            except PermissionError:
                # Evade processes owned by other users
                pass

        return None

    @staticmethod
    def get_qemu_pid():
        try:
            out = subprocess.check_output(["pgrep", "qemu-system"], encoding="utf8")
            pids = out.strip().split("\n")

            if len(pids) == 1:
                return int(pids[0], 10)
        except subprocess.CalledProcessError:
            # If no process with the name `qemu-system` is found, fallback to alternative methods,
            # as the binary name may vary (e.g., `qemu_system`).
            pass

        # We add a chardev file backend (we dont add a fronted, so it doesn't affect
        # the guest). We can then look through proc to find which process has the file
        # open. This approach is agnostic to namespaces (pid, network and mount).
        chardev_id = "gdb-pt-dump" + "-" + "".join(random.choices(string.ascii_letters, k=16))
        with tempfile.NamedTemporaryFile() as tmpf:
            pwndbg.dbg.selected_inferior().send_monitor(
                f"chardev-add file,id={chardev_id},path={tmpf.name}"
            )
            pid_found = QemuMachine.search_pids_for_file(pids, tmpf.name)
            pwndbg.dbg.selected_inferior().send_monitor(f"chardev-remove {chardev_id}")

        if not pid_found:
            raise ProcessLookupError("Could not find qemu-system pid")

        return int(pid_found, 10)

    def read_physical_memory(self, physical_address: int, length: int) -> bytes:
        res = pwndbg.dbg.selected_inferior().send_monitor(f"gpa2hva {hex(physical_address)}")

        # It's not possible to pread large sizes, so let's break the request
        # into a few smaller ones.
        max_block_size = 1024 * 1024 * 256
        try:
            hva = int(res.split(" ")[-1], 16)
            data = b""
            for offset in range(0, length, max_block_size):
                length_to_read = min(length - offset, max_block_size)
                block = os.pread(self.file, length_to_read, hva + offset)
                data += block
            return data
        except Exception as e:
            msg = f"Physical address ({hex(physical_address)}, +{hex(length)}) is not accessible. Reason: {e}. gpa2hva result: {res}"
            raise OSError(msg)

    def read_register(self, register_name: str) -> int:
        if register_name.startswith("$"):
            register_name = register_name[1:]

        return int(pwndbg.aglib.regs.read_reg(register_name))


@pwndbg.lib.cache.cache_until("stop")
def kernel_vmmap_via_page_tables() -> tuple[Page, ...]:
    if not pwndbg.aglib.qemu.is_qemu_kernel():
        return ()

    if sys.platform != "linux":
        # QemuMachine requires access to /proc/{qemu-pid}/mem, which is only available on Linux
        return ()

    try:
        machine_backend = QemuMachine()
    except PermissionError:
        print(
            message.error(
                "Permission error when attempting to parse page tables with gdb-pt-dump.\n"
                "Either change the kernel-vmmap setting, re-run GDB as root, or disable "
                "`ptrace_scope` (`echo 0 | sudo tee /proc/sys/kernel/yama/ptrace_scope`)"
            )
        )
        return ()
    except ProcessLookupError:
        print(
            message.error(
                "Could not find the PID for process named `qemu-system`.\n"
                "This might happen if pwndbg is running on a different machine than `qemu-system`,\n"
                "or if the `qemu-system` binary has a different name."
            )
        )
        return ()

    arch: str = pwndbg.aglib.arch.name
    ptrsize: int = pwndbg.aglib.arch.ptrsize
    if arch == "aarch64":
        arch_backend = PT_Aarch64_Backend(machine_backend)
    elif arch == "i386":
        arch_backend = PT_x86_64_Backend(machine_backend)
    elif arch == "x86-64":
        arch_backend = PT_x86_64_Backend(machine_backend)
    elif arch == "rv64":
        arch_backend = PT_RiscV64_Backend(machine_backend)
    else:
        print(
            message.error(
                f"The {pwndbg.aglib.arch.name} architecture does"
                " not support the `vmmap_via_page_tables`.\n"
                "Run `help show kernel-vmmap` for other options."
            )
        )
        return ()

    # If paging is not enabled, we shouldn't attempt to parse page tables
    if not pwndbg.aglib.kernel.paging_enabled():
        return ()

    p = PageTableDump(machine_backend, arch_backend)
    pages = p.arch_backend.parse_tables(p.cache, p.parser.parse_args(""))

    retpages: list[Page] = []
    for page in pages:
        start = page.va
        size = page.page_size
        flags = 4  # IMPLY ALWAYS READ
        if page.pwndbg_is_writeable():
            flags |= 2
        if page.pwndbg_is_executable():
            flags |= 1
        objfile = f"[pt_{hex(start)[2:-3]}]"
        retpages.append(Page(start, size, flags, 0, ptrsize, objfile))
    return tuple(retpages)


monitor_info_mem_not_warned = True


def _parser_mem_info_line_x86(line: str) -> Page | None:
    """
    Example response from `info mem`:
    ```
    ffff903580000000-ffff903580099000 0000000000099000 -rw
    ffff903580099000-ffff90358009b000 0000000000002000 -r-
    ffff90358009b000-ffff903582200000 0000000002165000 -rw
    ffff903582200000-ffff903582803000 0000000000603000 -r-
    ```
    """

    dash_idx = line.index("-")
    space_idx = line.index(" ")
    rspace_idx = line.rindex(" ")

    start = int(line[:dash_idx], 16)
    end = int(line[dash_idx + 1 : space_idx], 16)
    size = int(line[space_idx + 1 : rspace_idx], 16)
    perm = line[rspace_idx + 1 :]

    flags = 0
    if "r" in perm:
        flags |= Page.R_OK
    if "w" in perm:
        flags |= Page.W_OK
    if "x" in perm:
        flags |= Page.X_OK

    global monitor_info_mem_not_warned
    if end - start != size and monitor_info_mem_not_warned:
        print(
            message.warn(
                "The vmmap output may be incorrect as `monitor info mem` output assertion/assumption\n"
                "that end-start==size failed. The values are:\n"
                f"end={end:#x}; start={start:#x}; size={size:#x}; end-start={end - start:#x}\n"
                "Note that this warning will not show up again in this Pwndbg/GDB session."
            )
        )
        monitor_info_mem_not_warned = False

    return Page(start, size, flags, 0, pwndbg.aglib.arch.ptrsize, "<qemu>")


def _parser_mem_info_line_riscv64(line: str) -> Page | None:
    """
    Example response from `info mem`:
    ```
    vaddr            paddr            size             attr
    ---------------- ---------------- ---------------- -------
    0000000000010000 00000000feece000 0000000000001000 r-xu-a-
    0000000000011000 00000000fefeb000 0000000000002000 r-xu-a-
    0000000000013000 00000000a0a7a000 0000000000002000 r-xu-a-
    0000000000015000 00000000bfe02000 0000000000002000 r-xu-a-
    ```
    """

    arr = line.split(" ", 3)
    if len(arr) != 4:
        raise ValueError("invalid line format")

    start, _, size, perm = arr
    start = int(start, 16)
    size = int(size, 16)

    flags = 0
    if "r" in perm:
        flags |= Page.R_OK
    if "w" in perm:
        flags |= Page.W_OK
    if "x" in perm:
        flags |= Page.X_OK

    return Page(start, size, flags, 0, pwndbg.aglib.arch.ptrsize, "<qemu>")


@pwndbg.lib.cache.cache_until("stop")
def kernel_vmmap_via_monitor_info_mem() -> tuple[Page, ...]:
    """
    Returns Linux memory maps information by parsing `monitor info mem` output
    from QEMU kernel GDB stub.
    Works only on X86/X64/RISC-V as this is what QEMU supports.

    Consider using the `kernel_vmmap_via_page_tables` method
    as it is probably more reliable/better.

    See also: https://github.com/pwndbg/pwndbg/pull/685
    (TODO: revisit with future QEMU versions)
    """
    if not pwndbg.aglib.qemu.is_qemu_kernel():
        return ()

    try:
        monitor_info_mem = pwndbg.dbg.selected_inferior().send_monitor("info mem")
    except pwndbg.dbg_mod.Error:
        # Exception should not happen in new qemu, can we clean up it?
        # Older versions of QEMU/GDB may throw `gdb.error: "monitor" command
        # not supported by this target`. Newer versions will not throw, but will
        # return a string starting with 'unknown command:'.
        monitor_info_mem = "unknown command"

    parser_func = None
    if pwndbg.aglib.arch.name in ("i386", "x86-64"):
        parser_func = _parser_mem_info_line_x86
    elif pwndbg.aglib.arch.name == "rv64":
        parser_func = _parser_mem_info_line_riscv64

    if parser_func is None or "unknown command" in monitor_info_mem:
        print(
            message.error(
                f"The {pwndbg.aglib.arch.name} architecture does"
                " not support the `monitor info mem` command.\n"
                "Run `help show kernel-vmmap` for other options."
            )
        )
        return ()

    pages: list[Page] = []
    for line in monitor_info_mem.splitlines():
        try:
            page = parser_func(line)
        except Exception:
            # invalid format
            continue
        pages.append(page)

    return tuple(pages)


kernel_vmmap_mode = pwndbg.config.add_param(
    "kernel-vmmap",
    "page-tables",
    "the method to get vmmap information when debugging via QEMU kernel",
    help_docstring="""\
Values explained:

+ `page-tables` - walk page tables to render vmmap
+ `pt-dump` - read /proc/$qemu-pid/mem to parse kernel page tables to render vmmap
+ `monitor` - use QEMU's `monitor info mem` to render vmmap
+ `none` - disable vmmap rendering; useful if rendering is particularly slow

Note that the page-tables method will require the QEMU kernel process to be on the same machine and within the same PID namespace. Running QEMU kernel and GDB in different Docker containers will not work. Consider running both containers with --pid=host (meaning they will see and so be able to interact with all processes on the machine).
""",
    param_class=pwndbg.lib.config.PARAM_ENUM,
    enum_sequence=["page-tables", "pt-dump", "monitor", "none"],
)


@pwndbg.lib.cache.cache_until("stop")
def kernel_vmmap_pages() -> tuple[Page, ...]:
    mode = str(kernel_vmmap_mode)
    arch_name = pwndbg.aglib.arch.name
    if mode == "page-tables" and arch_name not in ("x86-64", "aarch64"):
        # TODO: remove this by implementing `RiscvPagingInfo`, `RiscvOps`, etc
        print(
            message.warn(
                f"`kernel-vmmap = {mode}` unsupported for {arch_name}, defaulting to `monitor`"
            )
        )
        mode = "monitor"
    match mode:
        case "page-tables":
            # has the user set the pgd with kcurrent?
            # None if not which gets properly handled
            entry = pwndbg.commands.kcurrent.KCURRENT_PGD
            if pwndbg.aglib.memory.is_kernel(entry):
                entry = pwndbg.aglib.kernel.pagewalk(entry, virt=False).phys
            return pwndbg.aglib.kernel.pagescan(entry)
        case "pt-dump":
            return kernel_vmmap_via_page_tables()
        case "monitor":
            return kernel_vmmap_via_monitor_info_mem()
    return ()


def kernel_vmmap() -> tuple[pwndbg.lib.memory.Page, ...]:
    if not pwndbg.aglib.qemu.is_qemu_kernel():
        return ()

    if pwndbg.aglib.arch.name not in (
        "i386",
        "x86-64",
        "aarch64",
        "rv32",
        "rv64",
    ):
        return ()

    pages = kernel_vmmap_pages()
    kv = KernelVmmap(pages)
    if kernel_vmmap_mode == "monitor" and pwndbg.aglib.arch.name == "x86-64":
        # TODO: check version here when QEMU displays the x bit for x64
        # see: https://github.com/pwndbg/pwndbg/pull/3020#issuecomment-2914573242
        for page in pages:
            if page.objfile == kv.pi.ESPSTACK:
                continue
            entry = pwndbg.aglib.kernel.pagewalk(page.start).entry
            if entry and entry >> 63 == 0:
                page.flags |= 1

    return tuple(pages)
